
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const Aggregator_cxx_Version = 
"$Id: Aggregator.cxx,v 1.19 2001/03/20 22:43:03 kevinmc Exp $";

#include "Aggregator.hxx"
#include "TRIPErrorHandler.hxx"
#include "LSSessionManager.hxx"
#include "cpLog.h"


Aggregator::Aggregator()
{

}


Aggregator::~Aggregator( )
{
}

void
Aggregator::process(TRIPTrieCache& trieCache)  
{
    cpLog(LOG_DEBUG, "Aggregator::process");
    // get all Non Aggregated and Aggregated routes in TRIPTrieCache
    LSSessionManager& sessionMgr = LSSessionManager::instance();
    const ProtoCapabilityList& protoList = sessionMgr.getProtoCapabilityList();
    const AddrFamilyList& addrFamilyList = sessionMgr.getAddrFamilyList();
    for (AddrFamilyList::const_iterator addrItr = addrFamilyList.begin();
                                    addrItr != addrFamilyList.end(); addrItr++)
    {
        for (ProtoCapabilityList::const_iterator itr = protoList.begin();
                                            itr != protoList.end(); itr++)
        {
            TRIPProtoKey key((*itr), (*addrItr));
            Sptr <TRIPRouteSet> trieCacheRouteSet = trieCache.getRouteSet(key);
            if (trieCacheRouteSet->size() == 0)
            {
                continue;
            }
            for(TRIPRouteSet::const_iterator itr = trieCacheRouteSet->begin();
                                    itr != trieCacheRouteSet->end(); itr++)
            {
                if ((*itr)->isAggregatedResult()) continue;
                if (!(*itr)->isChanged(T_AGGR_BIT)) continue;
                Sptr<TRIPRouteElement> route = (*itr);
                route->unsetChanged(T_AGGR_BIT);
                route->setChanged(RouteChangeBitMask(T_OUTGOING_BIT 
                                                            | T_PROXY_BIT));
                string newPrefix;
                Sptr <TRIPRouteElement> parent = 0;
                TRIPRouteSet childSet;
                TRIPProtoKey key( route->getAppProtocol(), 
                                                      route->getAddrFamily());
                if (trieCache.isAggregatable(route->getPrefix(), newPrefix, 
                                                        parent, childSet, key))
                {
                    // if parent exists and it is not an aggregation result
                    // then do not aggregate because dont want to overwrite
                    // Non-aggregated or Aggregated routes
                    if (( parent != 0 ) && 
                                        (!( parent->isAggregatedResult() )))
                    {
                        continue;
                    }
                    try
                    {
                        aggregate(childSet, trieCache, route, newPrefix);
                    }
                    catch (TRIPBadDataException& badData)
                    {
                        cpLog(LOG_ERR, "Aggregation failed due to bad data",
                                            badData.getDescription().c_str());
                        assert(0);
                    }
                    catch (TRIPAggregationException& aggr)
                    {
                        if ( parent != 0 ) 
                        {
                            parent->setWithdrawn();
                            parent->setChanged(RouteChangeBitMask
                                            (T_OUTGOING_BIT | T_PROXY_BIT));
                        }
                        cpLog(LOG_DEBUG, "Aggregation failed  due to aggregation mismatch",
                                                aggr.getDescription().c_str());
                        deAggregate(childSet);
                    }
                }
                else // not aggregatable
                {
                    if (( parent != 0 ) && ( parent->isAggregatedResult() ))
                    {
                        parent->setWithdrawn();
                        parent->setChanged(RouteChangeBitMask(T_OUTGOING_BIT
                                                            | T_PROXY_BIT));
                        deAggregate(childSet);
                    }
                }
            }
        }
    }
}

void 
Aggregator::checkAggregationOK(Sptr<TRIPRouteElement> route, 
                                    Sptr <TRIPNextHopServer>& newNextHop, 
                                    Sptr <TRIPMultiExitDisc>& newMultiExitDisc,
                                    TRIPRouteSet& retSet)
                        throw (TRIPBadDataException&, TRIPAggregationException&)
{
    Sptr<TRIPNextHopServer> nextHop = 0;
    Sptr<TRIPMultiExitDisc> multiExitDisc = 0;
    Sptr<TRIPMultiExitDisc> oldMultiExitDisc = 0;
    oldMultiExitDisc.dynamicCast(route->getAttribute(TRIP_MULTI_EXIT_DISC));
    TRIPRouteSet::const_iterator itr;
    for(itr = retSet.begin(); itr != retSet.end(); itr++)
    {
        // check that nextHops all match
        nextHop.dynamicCast((*itr)->getAttribute(TRIP_NEXT_HOP_SERVER));
        if (!(nextHop.getPtr()))
        {
            cpLog(LOG_ALERT, "Bad NextHop Server");
            throw TRIPBadDataException("Bad NextHop Server", __FILE__,
                                             __LINE__, TRIP_ERR_UPDATE);
        }
        if (!(newNextHop.getPtr()))
        { 
            newNextHop = nextHop;
        }
        else if (!(nextHop->isEqualAttribute(*newNextHop)))
        {
            cpLog(LOG_DEBUG, "Route (%s) NextHop failed to aggregate", 
            route->getPrefix().c_str());
            throw TRIPAggregationException("NextHop Server", __FILE__,
                                        __LINE__, 0);
        }
        // check that all multiExitDiscs match
        multiExitDisc.dynamicCast((*itr)->getAttribute(TRIP_MULTI_EXIT_DISC));
        if (!(multiExitDisc.getPtr()))
        {
            if (oldMultiExitDisc.getPtr())
            {
                cpLog(LOG_DEBUG, "Route (%s) MultiExit failed to aggregate",
                                                route->getPrefix().c_str());
                throw TRIPAggregationException("MultiExit", __FILE__, 
                __LINE__, 0);
            }
        }
        else
        {
            if (!(oldMultiExitDisc.getPtr()))
            {
                cpLog(LOG_DEBUG, "Route (%s) MultiExit failed to aggregate",
                                                route->getPrefix().c_str());
                throw TRIPAggregationException("MultiExit", __FILE__, 
                                                                __LINE__, 0);
            }
            else if (!(multiExitDisc->isEqualAttribute(*oldMultiExitDisc)))
            {
                cpLog(LOG_DEBUG, "Route (%s) MultiExit failed to aggregate",
                                                route->getPrefix().c_str());
                throw TRIPAggregationException("MultiExit", __FILE__, 
                                                                __LINE__, 0);
            }
        }
    }
    if (oldMultiExitDisc.getPtr())
    {
        newMultiExitDisc = new TRIPMultiExitDisc(oldMultiExitDisc->getPrefNumber()); 
    }
    u_int16_t itad = newNextHop->getItad();
    TRIPData hopAddr = newNextHop->getHopAddress();
    newNextHop = new TRIPNextHopServer(itad, hopAddr);
}


void 
Aggregator::aggregate(TRIPRouteSet& retSet, TRIPTrieCache& trieCache, 
                            Sptr<TRIPRouteElement> route, string& newPrefix) 
                    throw (TRIPBadDataException&, TRIPAggregationException&)
{
    cpLog(LOG_DEBUG, "Aggregator::aggregate route with prefix (%s)", 
                                                route->getPrefix().c_str());
    // first check that next hop servers and multi exit discs for
    // routes to be aggregated match, otherwise do not aggregate
    assert(route.getPtr());
    
    // NextHopServer - mandatory Attribute
    Sptr<TRIPNextHopServer> newNextHop = 0;
    // MultiExitDisc - non mandatory Attribute
    Sptr<TRIPMultiExitDisc> newMultiExit = 0;
    checkAggregationOK(route, newNextHop, newMultiExit, retSet);
    assert(newNextHop.getPtr());


    // Adv Path & Routed Path & Atomic Agg
    // Adv Path - mandatory Attribute
    Sptr<TRIPAdvertisePath> advPath;
    Sptr<TRIPAdvertisePath> newAdvPath = new TRIPAdvertisePath();

    // Routed Path - mandatory Attribute
    Sptr<TRIPRoutedPath> routedPath;
    Sptr<TRIPRoutedPath> newRoutedPath = new TRIPRoutedPath();

    // Atomic Agg - non mandatory Attribute
    Sptr<TRIPAtomicAggregate> atomicAgg;
    Sptr<TRIPAtomicAggregate> newAtomicAgg = 0;

    for(TRIPRouteSet::iterator itr = retSet.begin();itr != retSet.end(); itr++)
    {
        // adv path
        advPath.dynamicCast((*itr)->getAttribute(TRIP_ADVERTISEMENT_PATH));
        if (advPath.getPtr())
        {
            newAdvPath->aggregate(advPath);
        }

        // routed path
        routedPath.dynamicCast((*itr)->getAttribute(TRIP_ROUTED_PATH));
        if (routedPath.getPtr())
        {
            newRoutedPath->aggregate(routedPath);
        }

        // new atomic agg is only set if any of the aggregated routes are
        // found to have an atomic agg attribute
        if (!(newAtomicAgg.getPtr()))
        {
            atomicAgg.dynamicCast((*itr)->getAttribute(TRIP_ATOMIC_AGGREGATE));
            if (atomicAgg.getPtr())
            {
                newAtomicAgg = new TRIPAtomicAggregate();
            }
        }

        // reset the Aggr BIT if required on each of the routes being
        // aggregated as they do not need to be processed again by Aggregator
        // Must set Outgoing and Proxy BITs for OutgoingPolicy and Proxies.
        // Note this only applies to the routes which have changed
        if ((*itr)->isChanged(T_AGGR_BIT) == true)
        {
            (*itr)->unsetChanged(T_AGGR_BIT);
            (*itr)->setChanged(RouteChangeBitMask(T_OUTGOING_BIT
                                                                | T_PROXY_BIT));        }
    }
    
    // Community - non mandatory Attribute
    // Community handled last as it is not aggregated if atomicAgg exists
    Sptr<TRIPCommunity> newCommunity = 0;
    if (!(newAtomicAgg.getPtr())) // only aggregate community if no atomic Agg
    {
        Sptr <TRIPCommunity> community;
        for(TRIPRouteSet::iterator itr = retSet.begin();
                                                    itr != retSet.end(); itr++)
        {
            community.dynamicCast((*itr)->getAttribute(TRIP_COMMUNITY));
            if (community.getPtr())
            {
                if (!(newCommunity.getPtr()))
                {
                    newCommunity = new TRIPCommunity(*community);
                }
                newCommunity->aggregate(community);
            }
        }
    }

    // Create the new TRIPRouteElement
    TRIPRoutes routes = TRIPRoutes(route->getRoute().getAddrFamily(), route->getRoute().getAppProtocol(), TRIPData(newPrefix));
    TRIPItadKey itadKey = LSSessionManager::instance().getTripItadKey();
    Data nextHopData(newNextHop->getHopAddress().str());
    Sptr <TRIPRouteElement> newRoute = new TRIPRouteElement(itadKey, routes,
                                        nextHopData);
    newRoute->addAttribute(newNextHop);
    newRoute->addAttribute(newAdvPath);
    newRoute->addAttribute(newRoutedPath);
    if (newAtomicAgg.getPtr())
        newRoute->addAttribute(newAtomicAgg);
    if (newMultiExit.getPtr())
        newRoute->addAttribute(newMultiExit);
    if (newCommunity.getPtr())
        newRoute->addAttribute(newCommunity); 
    newRoute->setAggrStatus(T_AGGREGATED_RES);
    newRoute->setChanged(RouteChangeBitMask(T_OUTGOING_BIT | T_PROXY_BIT));
    //unset UPDATE BIT as it has been set as default
    newRoute->unsetChanged(T_UPDATE_BIT);
    cpLog(LOG_DEBUG, "Aggregator::insert aggregated route with prefix (%s)", 
                                                newRoute->getPrefix().c_str());
    trieCache.insert(newRoute);
    for(TRIPRouteSet::iterator itr = retSet.begin();itr != retSet.end(); itr++)
    {
        (*itr)->setAggrStatus(T_AGGREGATED);
    }
}

void
Aggregator::deAggregate(TRIPRouteSet& retSet)
{
    for(TRIPRouteSet::const_iterator itr = retSet.begin();
                                                itr != retSet.end(); itr++)
    {
        if ((*itr)->isAggregated())
        {
            (*itr)->setAggrStatus(T_NON_AGGREGATED);
        }
        if ((*itr)->isChanged(T_AGGR_BIT) == true)
        {
            (*itr)->unsetChanged(T_AGGR_BIT);
            (*itr)->setChanged(RouteChangeBitMask(T_OUTGOING_BIT 
                                                            | T_PROXY_BIT));
        }
    }
}


