/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const IncomingPolicyAgent_cxx_Version = 
"$Id: IncomingPolicyAgent.cxx,v 1.37 2001/03/20 22:50:35 kevinmc Exp $";

#include "IncomingPolicyAgent.hxx"
#include "TRIPUpdateMsg.hxx"
#include "LSUtils.hxx"
#include "TRIBAgent.hxx"
#include "DecisionProxyAgent.hxx"
#include "LSSessionManager.hxx"
#include "cpLog.h"
#include "TimeTracker.hxx"

IncomingPolicyAgent* IncomingPolicyAgent::_instance = 0;

IncomingPolicyAgent&
IncomingPolicyAgent::instance()
{
    if (!_instance)
    {
        _instance = new IncomingPolicyAgent();
    }
    return (*_instance);
}

IncomingPolicyAgent::IncomingPolicyAgent():_floodingFunc(0)
{
    cpLog(LOG_DEBUG, "IncomingPolicyAgent::IncomingPolicyAgent");
}

void
IncomingPolicyAgent::process(Sptr <TRIPUpdateMsg> update, 
                            const TRIPItadKey& key, Sptr <TRIPWorkSet> workSet,
                            TRIPProcessReason reason)
{
    cpLog(LOG_DEBUG, "IncomingPolicyAgent::process for Cache (%s)",
                                            key.getDescription().c_str());

    if (!(workSet.getPtr()))
    {
        assert(0);
    }

    if (reason == TRIP_CONNECTION_LOST)
    {
        handleLostConnection(key, workSet);
        return;
    }
    //Otherwise process the update msg
    assert(update.getPtr());
    Sptr <TRIPRouteSet> routeSet = LSUtils::translateRoutes(key, *update);
    if (routeSet->size() == 0) //if no elements, then nothing to process
    {
        cpLog(LOG_DEBUG, "IncomingPolicyAgent -no elements in Update from (%s)",
                                                key.getDescription().c_str());
        return;
    }

    bool internal = false;
    bool tribInUpdated = false;
    TRIBAgent& tribAgent = TRIBAgent::instance();
    tribAgent.writeLock();
    Sptr<TRIPTrieCache> adjTribIn;
    try
    {
        adjTribIn = tribAgent.getTribIn(key);
    }
    catch (VMissingDataException e)
    {
        cpLog(LOG_INFO, "IncomingPolicyAgent -no adj-TribIn exists for (%s)",                                                key.getDescription().c_str());
        tribAgent.unlock();
        return;
    }
    if (adjTribIn->isInternalRouteCache())
    {
        cpLog(LOG_DEBUG, "IncomingPolicyAgent - Internal Trib In Is Updated");
        internal = true;
    }
    else
    {
        cpLog(LOG_DEBUG, "IncomingPolicyAgent - External/Local Trib In Is Updated");
    }
    cpLog(LOG_INFO, "IncomingPolicyAgent - number of elements is (%d)", 
                                                        routeSet->size());
    for(TRIPRouteSet::iterator itr = routeSet->begin();
                                            itr != routeSet->end(); itr++)
    {
        Sptr<TRIPRouteElement> route = (*itr);
        if (route->isWithdrawn()) //Withdrawn Route
        {
            tribInUpdated |= handleWithdrawnRoute(route, adjTribIn, 
                                                    internal, workSet);
        }
        else // Reachable Route
        {
            tribInUpdated |= handleReachableRoute(route, adjTribIn, 
                                                    internal, workSet);
        }
    }
    if (tribInUpdated == true)
    {
        adjTribIn->setUpdated();
    }
    //The internal adjTribIn with Local key contains all the routes 
    //originated by this LS which have already been processed by DecisionAgent
    //These routes should still be flooded.
    if ((internal == true) && (key.isLocal()))
    {
        DecisionProxyAgent::instance().setDecisionOn(false);
        adjTribIn->setUpdated(false);
    }

    if ((internal == true) && (tribInUpdated == true))
    {
        // Call the Call Back Fn for Flooding
        if (_floodingFunc != 0)
        {
            cpLog(LOG_DEBUG, "Calling flooding callback function for (%s)", key.getDescription().c_str());
            _floodingFunc(workSet, TRIP_CB_INTERNAL_UPDATE, key);
        }
    }
    tribAgent.unlock();
}


bool
IncomingPolicyAgent::internalRouteIsValid(const Sptr<TRIPRouteElement> route,
                            Sptr<TRIPTrieCache> tribIn, bool& matchedDest)
{
    cpLog(LOG_DEBUG, "IncomingPolicyAgent::internalRouteIsValid");
    TRIPRouteSet retSet;
    if (tribIn->matchDestination(route, retSet))
    {
        // if more than one matching in adjTribIn => error
        if (retSet.size() > 1)
        {
            cpLog(LOG_ALERT, "More than one matching route, number matching = (%d)", retSet.size());
            assert(0);
        }
        matchedDest = true;
        Sptr<TRIPRouteElement> matching = *(retSet.begin());
        if (route->getSeqNumber() <= matching->getSeqNumber())
        {
            return false;
        }
    }
    return true;
}

    
bool
IncomingPolicyAgent::handleWithdrawnRoute(const Sptr<TRIPRouteElement> newRoute,
                                    Sptr<TRIPTrieCache> tribIn,
                                    bool isInternal, Sptr <TRIPWorkSet> workSet)
{
    cpLog(LOG_DEBUG, "IncomingPolicyAgent::handleWithdrawnRoute");
    bool updating = false;
    TRIPRouteSet retSet;
    if (tribIn->matchDestination(newRoute, retSet))
    {
        if (retSet.size() > 1)
        {
            cpLog(LOG_ALERT, "More than one matching route, number matching = (%d)", retSet.size());
        }
        Sptr<TRIPRouteElement> oldRoute = *(retSet.begin());

        if (isInternal == true)
        {
            if (newRoute->getSeqNumber() > oldRoute->getSeqNumber())
            {
                updating = true;
                oldRoute->setChanged(RouteChangeBitMask(T_FLOOD_BIT)); 
                oldRoute->setSeqNumber(newRoute->getSeqNumber());
                oldRoute->setTransactionId(newRoute->getTransactionId());
            }
        }
        else //external
        {
            updating = true;
        }
        if (updating)
        {
            //tribIn->insert(oldRoute);
            oldRoute->setWithdrawn();
            oldRoute->setChanged(RouteChangeBitMask(T_UPDATE_BIT));
            workSet->insert(oldRoute);
            DecisionProxyAgent::instance().setDecisionOn(true);
        }
    }
    else
    {
        cpLog(LOG_ERR, "No match for Withdrawn Route in Cache (%s)",
        tribIn->getItadKey().getDescription().c_str());
    }
    return updating;
}


bool
IncomingPolicyAgent::handleReachableRoute(const Sptr<TRIPRouteElement> newRoute,
                                    Sptr<TRIPTrieCache> tribIn,
                                    bool isInternal, Sptr <TRIPWorkSet> workSet)
{
    cpLog(LOG_DEBUG, "IncomingPolicyAgent::handleReachableRoute");
    bool updating = false;
    bool matchedDest = false;
    TRIPRouteSet retSet;
    if (isInternal)
    {
        if (internalRouteIsValid(newRoute, tribIn, matchedDest))
        {
            updating = true;
            newRoute->setChanged(RouteChangeBitMask(T_FLOOD_BIT));
        }
    }
    else //external
    {
        u_int32_t prefValue = 0;
        updating = true;
        prefValue = calculatePref(*newRoute);
        matchedDest = tribIn->matchDestination(newRoute, retSet);
        // set Localpref
        Sptr<TRIPLocalPref> localPref;
        localPref.dynamicCast(newRoute->getAttribute(TRIP_LOCAL_PREF));
        if (!(localPref.getPtr()))
        {
            localPref = new TRIPLocalPref(prefValue);
            newRoute->addAttribute(localPref);
        }
        localPref->setPrefNumber(prefValue);
    }

    if (updating)
    {
        bool sameAttr = false;
        if (matchedDest)
        {
            cpLog(LOG_DEBUG, "IncomingPolicyAgent - matching route exists");
            DecisionProxyAgent::instance().setDecisionOn(true);
        }
        else if (tribIn->moreSpecific(newRoute, retSet, sameAttr))
        {
            if(!sameAttr)
            {
                cpLog(LOG_DEBUG, "IncomingPolicyAgent - more spec not same attr");
                if (retSet.size() > 1)
                {
                    cpLog(LOG_ALERT, "More than one matching route, number matching = (%d)", retSet.size());
                }
                // mark the more specific, not same attr route as changed
                Sptr<TRIPRouteElement> oldRoute = *(retSet.begin());
                oldRoute->setChanged(T_UPDATE_BIT);
                DecisionProxyAgent::instance().setDecisionOn(true);
            }
            else
            {
                cpLog(LOG_DEBUG, "IncomingPolicyAgent - more spec. same attr");
            }
        }
        else // otherwise does not exist or is less specific
        {
            cpLog(LOG_DEBUG, "IncomingPolicyAgent - less spec. or no route ");
            DecisionProxyAgent::instance().setDecisionOn(true);
        }
        // add the new route
        tribIn->insert(newRoute);
        workSet->insert(newRoute);
    }
    return updating;
}
    
void
IncomingPolicyAgent::init()
{
    //TODO Will be used for Policy data
}

u_int32_t
IncomingPolicyAgent::calculatePref(const TRIPRouteElement& route)
{
    //TODO Policy Data to be plugged in to route for pref value
    //Temporarily returning a value between 1 & 20 based on the TRIPId
    return (((route.getSource().getTripId().getId())%20)+1);
}

void
IncomingPolicyAgent::handleLostConnection(const TRIPItadKey& key, 
                                                    Sptr <TRIPWorkSet> workSet)
{
    cpLog(LOG_DEBUG, "IncomingPolicyAgent::handleLostConnection from (%s)",                                                key.getDescription().c_str());
    TRIBAgent& tribAgent = TRIBAgent::instance();
    tribAgent.writeLock();
    Sptr<TRIPTrieCache> adjTribIn;
    try
    {
        adjTribIn = tribAgent.getTribIn(key);
    }
    catch (VMissingDataException e)
    {
        cpLog(LOG_INFO, "IncomingPolicyAgent -no adj-TribIn exists for (%s)",                                                key.getDescription().c_str());
        tribAgent.unlock();
        return;
    }
    LSSessionManager& mgr = LSSessionManager::instance();
    const ProtoCapabilityList& pList = mgr.getProtoCapabilityList();
    const AddrFamilyList& aList = mgr.getAddrFamilyList();
    for (AddrFamilyList::const_iterator adItr = aList.begin();
                                                adItr != aList.end(); adItr++)
    {
        for (ProtoCapabilityList::const_iterator pItr = pList.begin();
                                            pItr != pList.end(); pItr++)
        {
            TRIPProtoKey pkey((*pItr), (*adItr));
            Sptr<TRIPRouteSet> routeSet = adjTribIn->getRouteSet(pkey);
            for (TRIPRouteSet::const_iterator itr = routeSet->begin();
                                            itr != routeSet->end(); itr++)
            {
                (*itr)->setWithdrawn();
                (*itr)->setChanged(RouteChangeBitMask(T_UPDATE_BIT));
                workSet->insert((*itr));
            }
        }
    }
    if (workSet->getRouteSet().size())
    {
        DecisionProxyAgent::instance().setDecisionOn(true);    
        adjTribIn->setUpdated();
    }
    tribAgent.unlock();

}
