/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSPeerSession_cxx_Version = 
"$Id: LSPeerSession.cxx,v 1.49 2001/03/20 22:55:03 kevinmc Exp $";

#include <iostream.h>
#include <algorithm>
#include <strstream.h>
#include "LSPeerSession.hxx"
#include "TcpClientSocket.hxx"
#include "TRIPTransceiver.hxx"
#include "TRIPOpenMsg.hxx"
#include "TRIPUpdateMsg.hxx"
#include "TRIPStateUnknown.hxx"
#include "TRIPNotificationMsg.hxx"
#include "TRIPKeepAliveMsg.hxx"
#include "LSLocator.hxx"
#include "LockHelper.hxx"
#include "LSSessionManager.hxx"
#include "DecisionProxyAgent.hxx"
#include "KeepAliveAgent.hxx"
#include "HoldTimeAgent.hxx"
#include "RestartAgent.hxx"
#include "ConnectRetryAgent.hxx"
#include "OutgoingPolicyAgent.hxx"
#include "TRIBAgent.hxx"
#include "TRIPDef.hxx"
#include "VCondition.h"
#include "cpLog.h"

LSPeerSession::LSPeerSession(u_int16_t itad,
                 const TRIPId& tripId,const NetworkAddress& lsAddress)
{
    _shutdown = false;
    _itadKey = TRIPItadKey(itad, tripId);
    _peerAddress = new NetworkAddress(lsAddress);
    _destIp = _peerAddress->getIpName();

    _state = new TRIPStateUnknown();
    _state->init(*this);
    _thread = new VThread();
    _thread->spawn(LSPeerSession::run, this);
}

LSPeerSession::LSPeerSession(u_int16_t itad, u_int16_t pItad,
                      const TRIPId& tripId, const TRIPId& pTripId,                                             const Connection& conn)
{
    _shutdown = false;
    _itadKey = TRIPItadKey(itad, tripId);
    _peerItadKey = TRIPItadKey(pItad, pTripId);
    _connection = conn;
    _destIp = _connection.getIp();

    _state = new TRIPStateUnknown();
    _state->init(*this);

    _thread = new VThread();
    _thread->spawn(LSPeerSession::run, this);
}

void
LSPeerSession::createOpenMsg()
{
    //Set the default HoldTimer to be 4 minutes
    _openMsg = new TRIPOpenMsg(_itadKey.getItad(), _itadKey.getTripId().getId(), 4*60);
    Sptr<TRIPCapabilityInfo>  cInfo = new TRIPCapabilityInfo();

    Sptr<TRIPTransCapability> cap = new TRIPTransCapability(getMode()); 
    cInfo->addCapability(cap);

    Sptr<TRIPRouteCapability> rCap = new TRIPRouteCapability();
    const ProtoCapabilityList& pList = 
                  LSSessionManager::instance().getProtoCapabilityList();
    for(ProtoCapabilityList::const_iterator itr = pList.begin(); 
                   itr != pList.end(); itr++)
    {
        Sptr<TRIPRouteHdr> rHdr1 = new TRIPRouteHdr(TRIP_ADDR_POTS, (*itr));
        rCap->addRouteCapability(rHdr1);
    }
    cInfo->addCapability(rCap);

    _openMsg->addParameter(cInfo);
}

LSPeerSession::~LSPeerSession()
{
    _shutdown = true;
    releaseTRIPResources();
    if(_connection.isLive()) _connection.close();
    if(_thread.getPtr() != 0)
    {
        _thread->join();
        _thread->exit();
    }
}

void
LSPeerSession::start()
{
    LockHelper lck(_mutex);
    cpLog(LOG_DEBUG, "LSPeerSession::start");
    try
    {
        _state->start(*this);
        connect();
        _state->sendOpen(*this);
    }
    catch(VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Network error, reason:%s",
                          e.getDescription().c_str());
        _state->lostConnection(*this);
    }
}


void
LSPeerSession::connect()
{
    if(_connection.isLive())
    {
        cpLog(LOG_INFO, "Already connected to (%s)", 
                                 _connection.getDescription().c_str());        
        return;
    }

    //Create a Client Socket.When ClientSocket destroys, it also
    //close the connection.Passing false in the constructor , tells
    //object not to close the connection.
    TcpClientSocket cSock(*_peerAddress, false);
   
    cSock.connect();
    _connection = cSock.getConn();

    assert(_connection.isLive());
    cpLog(LOG_INFO, "Connected to (%s)", _connection.getDescription().c_str());
}

void 
LSPeerSession::dump(ostream& os) const
{
    os << "(" << _itadKey <<  ")<->(" << _peerItadKey <<  ")" << endl; 
    os << "(" << LSSessionManager::instance().getLocalIp().c_str() <<  ")<->(" << _destIp.c_str() <<  ")" << endl; 
}

bool 
LSPeerSession::peerExists(u_int16_t itad, const TRIPId& tripId)
{
    return((_peerItadKey.getItad() == itad) && 
       (_peerItadKey.getTripId() == tripId));
}

bool 
LSPeerSession::peerExists(const string& destIp)
{
    return(_destIp == destIp);
}

void*
LSPeerSession::run(void* args)
{
    LSPeerSession* self = static_cast<LSPeerSession*>(args);
    self->processMsg();
    return (0);
}

void
LSPeerSession::processMsg()
{
    TRIPTransceiver trans;
    for (; ; )
    {
        LockHelper lck(_mutex);
        if(_connection.isLive() && _connection.isReadReady(1))
        {
            try
            {
                Sptr<TRIPMsg> msg = trans.receiveMsg(_connection.getConnId());
                assert(msg != 0);
                switch(msg->getType())
                {
                    case TRIP_OP_KEEPALIVE:
                    {
                        _state->receiveKeepAlive(*this);
                        ostrstream oStr;
                        oStr << *this << ends;
                        cpLog(LOG_DEBUG, "Received KeepAlive\n %s", oStr.str());
                        oStr.freeze(false);
                        if (_state->needsDataSync() && 
                            _state->isEstablished())
                        {
                            cpLog(LOG_INFO, "Session established (%s)->(%s)",
                                            LSSessionManager::instance().getLocalIp().c_str(), _destIp.c_str());
                            dataSync();
                            //Also send the topology update
                            if(_peerItadKey.isInternal())
                            {
                                LSSessionManager::instance().addTopology(_itadKey, _peerItadKey);
                            }
                        }
                    }
                    break;
                    case TRIP_OP_UPDATE:
                        handleUpdateMessage(msg);
                    break;
                    case TRIP_OP_OPEN:
                        processOpenMsg(msg);
                    break;
                    case TRIP_OP_NOTIF:
                        _state->receiveNotification(*this);
                    break;
                    default:
                       throw TRIPBadDataException("Unupported message type",
                                        __FILE__, __LINE__, TRIP_ERR_OPEN, 0);
                    break;
                }
            }
            catch(VNetworkException& e)
            {
                cpLog(LOG_ALERT, "Network error, reason:%s",
                                     e.getDescription().c_str());
                _state->lostConnection(*this);
            }
            catch(TRIPInvalidStateException& e)
            {
                cpLog(LOG_ALERT, "Invalid state transition , reason:%s",
                                     e.getDescription().c_str());
                assert(0);
            }
            catch(TRIPBadDataException& e)
            {
                cpLog(LOG_ALERT, "Bad TRIP data, reason:%s",
                                     e.getDescription().c_str());
                if(_state->isEstablished())
                {
                    sendNotification(e.getError(), e.getErrSubcode(), e.getMsg().c_str()); 
                }
                _state->lostConnection(*this);
            }
        }
        else if(!_connection.isLive())
        {
            //Unlock the lock , before sleeping.The unlock
            //will be called again when LockHelper scope ends, 
            //however the multiple call to unlock will be no-op.
            lck.unlock();
            vusleep(100000);
        }
        if(_shutdown) 
        {
             break;
        }
    }
}

void
LSPeerSession::processOpenMsg(Sptr<TRIPMsg> msg) throw (VNetworkException&, TRIPBadDataException&)
{
    cpLog(LOG_DEBUG, "processOpenMsg (%s)<->(%s)", LSSessionManager::instance().getLocalIp().c_str(), _destIp.c_str());

    if(msg == 0) 
    {
         assert(0);
         return;
    }

    Sptr<TRIPOpenMsg> openMsg;
    openMsg.dynamicCast(msg);
    assert(openMsg != 0);
    _peerItadKey = TRIPItadKey(openMsg->getItad(), openMsg->getTripId());
    Sptr<TRIPTransCapability> transCap = openMsg->getTransCapability();
    assert(transCap != 0);
    if((transCap->isSendOnly() && (getMode() == TRIP_TRANS_SND)) ||
       (transCap->isRecvOnly() && (getMode() == TRIP_TRANS_RCV)))
    {
        //Send the notification and close the connection
        cpLog(LOG_ALERT, "Unsupported capability");
        TRIPData data("Unsupported capability");
        TRIPNotificationMsg msg(TRIP_ERR_OPEN, O_UNSUPP_CAPABILITY, data);
        LockHelper lck(_sendLock);
        msg.writeData(_connection.getConnId());
        _state->fail(*this);
        return;
    }
    try
    {
        _state->receiveOpen(*this);
        sendKeepAlive();
        //The hold timer value comes as second, so convert it into
        //milliseconds and set it.Also the keep alive timer value
        //is 1/3 of the hold timer value, also set in milliseconds
        _holdTimeAgent->setDuration(openMsg->getHoldTime()*1000);
        _keepAliveAgent->setDuration((openMsg->getHoldTime()*1000)/3);
    }
    catch(TRIPInvalidStateException& e)
    {
        cpLog(LOG_DEBUG, "Invalid state transition, (%s)",
                         e.getDescription().c_str());
        assert(0);
    }
    catch(VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Network error, (%s)",
                         e.getDescription().c_str());
        _state->fail(*this);
    }
}

ostream&
operator<<(ostream& os, const LSPeerSession& lsSession)
{
    lsSession.dump(os);
    return os;
}

void
LSPeerSession::sendNotification(u_int8_t eCode, u_int8_t eSubCode, const string& msg)
{
    try
    {
        TRIPData data(msg);
        TRIPNotificationMsg msg(eCode, eSubCode, data);
        if(_connection.isLive())
        {
            LockHelper lck(_sendLock);
            msg.writeData(_connection.getConnId());
        }
    }
    catch(VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Failed to send notification msg");
    }
}

void
LSPeerSession::sendKeepAlive()
{
    if (!_shutdown)
    {
        cpLog(LOG_DEBUG, "sendKeepAlive (%s)<->(%s)", LSSessionManager::instance().getLocalIp().c_str(), _destIp.c_str());
        try
        {
            TRIPKeepAliveMsg msg;
            LockHelper lck(_sendLock);
            msg.writeData(_connection.getConnId());
        }
        catch(VException& e)
        {
            cpLog(LOG_INFO, "Got exception (%s), in state (%s)",
                    e.getDescription().c_str(), _state->description().c_str());
            _state->fail(*this);
        }
        TRIBAgent::instance().prettyDump(cerr);
    }
}

void
LSPeerSession::dataSync()
{
    cpLog(LOG_INFO, "Initial connection, synchronizing databases...");
    if (isPeerInternal())
    {
        TRIBAgent& tribAgent = TRIBAgent::instance();
        LSSessionManager& mgr = LSSessionManager::instance();
        const ProtoCapabilityList& pList = mgr.getProtoCapabilityList();
        const AddrFamilyList& aList = mgr.getAddrFamilyList();
        Sptr<TRIPTrieList > tribsToFlood = tribAgent.getAllInternalTribIns();
        Sptr<TRIPTrieCache> extTrib = tribAgent.getExtTrib();
        tribsToFlood->push_back(extTrib);
        for(TRIPTrieList::iterator itr = tribsToFlood->begin(); 
                                            itr != tribsToFlood->end(); itr++)
        {
            (*itr)->readLock();
            for (AddrFamilyList::const_iterator adItr = aList.begin();
                                    adItr != aList.end(); adItr++)
            {
                for (ProtoCapabilityList::const_iterator pItr = pList.begin();
                                    pItr != pList.end(); pItr++)
                {
                    TRIPProtoKey pkey((*pItr), (*adItr));
                    Sptr<TRIPRouteSet> floodSet = (*itr)->getSortedRouteSet(pkey);
                    if (floodSet.getPtr())
                    {
                        Sptr <UpdateMsgList> msgList = mgr.getFloodMsgList(*floodSet, 
                                    true, !(*itr)->isInternalRouteCache());
                        if (msgList.getPtr())
                        {
                            sendUpdateMsgList(msgList);
                        }
                    }
                }
            }
            (*itr)->unlock();
        }
    }
    else
    {
        OutgoingPolicyAgent::instance().addNewPeerKey(getPeerItadKey());
    }
    cpLog(LOG_INFO, "Synchronizing done...");
}

void
LSPeerSession::sendUpdateMsgList(Sptr <UpdateMsgList> msgList)
{
    if (!(msgList.getPtr()))
    {
        return;
    }
    for (UpdateMsgList::const_iterator itr = msgList->begin();
                                        itr != msgList->end(); itr++)
    {
        sendUpdate(*(*itr));
    }
}

void
LSPeerSession::holdTimerExpired() 
{
    if (!_shutdown)
    {
        _state->holdTimerExpired(*this);
    }
}

void
LSPeerSession::restartTimerExpired() 
{
    if (!_shutdown)
    {
        cpLog(LOG_DEBUG, "LSPeerSession::restartTimerExpired");
        LockHelper lck(_mutex);
        try
        {
            _state->start(*this);
            connect();
            _state->sendOpen(*this);
        }
        catch(VNetworkException &e)
        {
            cpLog(LOG_DEBUG, "Network error, reason:%s",
                                e.getDescription().c_str());
            _state->lostConnection(*this);
        }
    }
}

void
LSPeerSession::makeConnection()
{
    setConnectRetryTimer();
    LockHelper lck(_mutex);
    connect();
}

void
LSPeerSession::closeConnections()
{
    if(_connection.isLive())
    {
        _connection.close();
    }
}

void
LSPeerSession::initTRIPResources()
{
    // TODO Initialize required resources
    // Need to create the Timer Agents
    // Need to set all of the initial timer durations if required.
    _keepAliveAgent = new KeepAliveAgent(*this);
    _holdTimeAgent = new HoldTimeAgent(*this);
    _connectRetryAgent = new ConnectRetryAgent(*this);
    _restartAgent = new RestartAgent(*this);

    // Initial timer durations - may be changed later after peer negotiation
    _keepAliveAgent->setDuration(DEFAULT_KEEPALIVE_TIME); //1/3 of holdTimer
    // recommended initial 4 mins
    _holdTimeAgent->setDuration(DEFAULT_HOLD_TIME); 
    // locally set time - 1/2 min
    _connectRetryAgent->setDuration(DEFAULT_CONN_RETRY_TIME); 
    // recommended 1 min initially
    _restartAgent->setDuration(DEFAULT_RESTART_TIME); 
}

void
LSPeerSession::releaseTRIPResources()
{
    _keepAliveAgent->cancelTimer();
    _holdTimeAgent->cancelTimer();
    _connectRetryAgent->cancelTimer();
    _restartAgent->cancelTimer();
    // reset to initial value
    _connectRetryAgent->setDuration(DEFAULT_CONN_RETRY_TIME); 
    LSSessionManager::instance().clearAdjTribIn(_peerItadKey);
}

void 
LSPeerSession::handleIncomingConnection(Connection& conn) 
                               throw (VNetworkException&)
{
    LockHelper lck(_mutex);
    if(_state->acceptAllowed())
    {
        LSLocator lslocator;
        if(!LSSessionManager::instance().getLsLocator().isTrustedLs(conn.getIp()))
        {
            cpLog(LOG_DEBUG, "LS is not trusted, ignoring the connection");
            conn.close();
            return;
        }
        cpLog(LOG_DEBUG, "LSPeerSession::handleIncomingConnection, acceptAllowed");
        if(_connection.isLive())
        {
            //At this point LS initiated an OPEN connection and at the same
            //time got the OPEN request from the Peer, so we close our
            //connection and start using peer's.
            cpLog(LOG_INFO, "Closing connection (%s)", _connection.getDescription().c_str());
            _connection.close();
        }
        _connection = conn;
        cpLog(LOG_DEBUG, "Assigned connection (%s)", _connection.getDescription().c_str());
        _state->sendOpen(*this);
    }
    else
    {
        //Session is not in state of accepting OPEN connection
        cpLog(LOG_DEBUG, "LSPeerSession::handleIncomingConnection, accept not Allowed");
        cpLog(LOG_DEBUG, "Closing connection (%s)", conn.getDescription().c_str());
        conn.close();
    }
}

void
LSPeerSession::handleUpdateMessage(Sptr<TRIPMsg> msg)
{
    cpLog(LOG_DEBUG, "LSPeerSession::handleUpdateMessage");
    //Can only handle update message in established state
    if (!(_state->isEstablished()))
    {
        cpLog(LOG_INFO, "LSPeerSession::handleUpdateMessage - session not Established for (%s), ignoring Update Msg", 
                                _peerItadKey.getDescription().c_str());
        return;
    }

    Sptr<TRIPUpdateMsg> uMsg;
    uMsg.dynamicCast(msg);
    Sptr<TRIPItadKey> originItadKey = 0;
    if(uMsg.getPtr() != 0)
    {
        ///Validate the Update message first
        u_int8_t eCode, subCode;
        string emsg;
        if(!uMsg->validate(LSSessionManager::instance().getItad(), eCode, subCode, emsg))
        {
            cpLog(LOG_INFO, "Failed to validate the Update Msg. Subcode (%d)", subCode);
            sendNotification(eCode, subCode, emsg);
            _state->lostConnection(*this);
            return;
        }
        if(isPeerInternal())
        {
           //Check that ReachanbleRoute, WithdrawnRoute and ITAD Topology
           //Attributes are LinkState attribute
           if(!uMsg->checkLinkStateAttributes(eCode, subCode, emsg))
           {
                cpLog(LOG_INFO, "Failed to validate Link-State Attr in the Update Msg. Subcode (%d)", subCode);
                sendNotification(eCode, subCode, emsg);
                _state->lostConnection(*this);
                return;
           }
           ///Get the ITAD topology attribute, if it has one
           if(uMsg->hasItadTopology())
           {
               Sptr<TRIPRouteAttr> rAttr = uMsg->getAttribute(TRIP_ITAD_TOPOLOGY);
               Sptr<TRIPItadTopology> itadTopology;
               itadTopology.dynamicCast(rAttr);
               assert(itadTopology != 0);
               LSSessionManager::instance().handleTopologyChanges(_peerItadKey,itadTopology);
           }
           //Get the originator TRIPId from the Link-state attribute
           //and form the ItadKey for originator, since it is internal
           //it has the same itad as the peer
           const Sptr<TRIPWithdrawnRoute> wRoute =  uMsg->getWithdrawnRoute(); 
           originItadKey = new TRIPItadKey(_peerItadKey.getItad(),
                                           wRoute->getTripId());
           TRIBAgent::instance().createAdjTribIn(*originItadKey, TC_INTERNAL);
        }
        else
        {
            //There should not be link state attributes in External message 
            if(uMsg->extCheckInvalidAttributes(eCode, subCode, emsg))
            {
                cpLog(LOG_INFO, "Link-State Attr in the Update Msg. Subcode (%d)", subCode);
                sendNotification(eCode, subCode, emsg);
                _state->lostConnection(*this);
                return;
            }
            originItadKey = new TRIPItadKey(_peerItadKey);
            TRIBAgent::instance().createAdjTribIn(_peerItadKey, TC_EXTERNAL);
        }
      
        //Assign a transactionID to the update message 
        uMsg->setTransactionId(LSSessionManager::instance().getNextTransactionId());
        cpLog(LOG_DEBUG, "Putting in the incoming Fifo"); 
        Sptr<TRIPUpdateEvent> event = new TRIPUpdateEvent(uMsg, *originItadKey);
        DecisionProxyAgent::instance().addToQueue(event);
    }

}

void
LSPeerSession::advertiseRoutes()
{
    //TESTING ONLY
    //Create the UPdate Message
    TRIPUpdateMsg uMsg;
    Sptr<TRIPWithdrawnRoute> route1 = new TRIPWithdrawnRoute();
    Sptr<TRIPReachableRoute> route2 = new TRIPReachableRoute();
    TRIPRoutes tRoute(TRIP_ADDR_POTS, TRIP_APP_PROTO_SIP,TRIPData("1408"));
    TRIPRoutes tRoute2(TRIP_ADDR_POTS, TRIP_APP_PROTO_SIP,TRIPData("1508"));
    *route1 << tRoute;
    *route2 << tRoute2;

    uMsg.addWithdrawnRoute(route1);
    uMsg.addReachableRoute(route2);
  
    TRIPItadPathSegment seg1(TRIP_SEG_AP_SET, _itadKey.getItad());
    Sptr<TRIPAdvertisePath> aPath = new TRIPAdvertisePath();
    *aPath << seg1;
    
    TRIPItadPathSegment seg2(TRIP_SEG_AP_SET, _itadKey.getItad());
    Sptr<TRIPRoutedPath> aPath2 = new TRIPRoutedPath();
    *aPath2 << seg2;

    uMsg.addAttribute(aPath); 
    uMsg.addAttribute(aPath2); 

    Sptr<TRIPNextHopServer> nServer = new TRIPNextHopServer(_itadKey.getItad(), 
                                                      TRIPData("129.234.34.56"));
    uMsg.addAttribute(nServer);
    cpLog(LOG_DEBUG,  "Sending Update Msg");
    sendUpdate(uMsg);
}

void 
LSPeerSession::sendUpdate(const TRIPUpdateMsg& uMsg) throw (VNetworkException&)
{
    if(_connection.isLive())
    {
        LockHelper lck(_sendLock);
        uMsg.writeData(_connection.getConnId());
    }
}


bool
LSPeerSession::isPeerInternal() const
{
    return(_itadKey.getItad() == _peerItadKey.getItad());
}

void
LSPeerSession::sendOpen() throw (VNetworkException&)
{

    assert(_connection.isLive());
    createOpenMsg();
    assert(_openMsg != 0);
    ///Send OPEN MSG
    cpLog(LOG_DEBUG, "Sending OPEN Msg (%s)<->(%s)", LSSessionManager::instance().getLocalIp().c_str(), _destIp.c_str());

    strstream oStr;
    oStr << *_openMsg << endl << ends;
    cpLog(LOG_DEBUG, "%s", oStr.str());
    oStr.freeze(false);
    LockHelper lck(_sendLock);
    _openMsg->writeData(_connection.getConnId());
}

void 
LSPeerSession::setHoldTimer()
{
    _holdTimeAgent->startTimer();
}

void 
LSPeerSession::setConnectRetryTimer()
{
    _connectRetryAgent->startTimer();
}

void 
LSPeerSession::initKeepAlive()
{
    _keepAliveAgent->startTimer();
}

void 
LSPeerSession::connectRetryTimerExpired()
{
    if (!_shutdown)
    {
        _state->connectRetryTimerExpired(*this);
    }
};

void 
LSPeerSession::cancelConnectRetryTimer()
{
    _connectRetryAgent->cancelTimer();
}

void 
LSPeerSession::cancelHoldTimer()
{
    _holdTimeAgent->cancelTimer();
}

void 
LSPeerSession::setRestartTimer()
{
    _restartAgent->startTimer();
}


void 
LSPeerSession::stopKeepAlive()
{
    _keepAliveAgent->cancelTimer();
}



