/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSSessionManager_cxx_Version = 
"$Id: LSSessionManager.cxx,v 1.45 2001/03/20 22:57:15 kevinmc Exp $";


#include <unistd.h>
#include <iostream>
#include <signal.h>
#include <algorithm>
#include <stdlib.h> 
#include <strstream.h>
#include "LSSessionManager.hxx"
#include "LSSendOnlySession.hxx"
#include "LSReceiveOnlySession.hxx"
#include "LSSendReceiveSession.hxx"
#include "TcpServerSocket.hxx"
#include "TRIPOpenMsg.hxx"
#include "TRIPTransceiver.hxx"
#include "TRIBAgent.hxx"
#include "OutgoingPolicyAgent.hxx"
#include "IncomingPolicyAgent.hxx"
#include "DecisionAgent.hxx"
#include "DecisionProxyAgent.hxx"
#include "LSApiServer.hxx"
#include "LSUtils.hxx"
#include "cpLog.h"

static const int LS_PORT=6069;
static const int LS_API_DEF_PORT=8000;

LSSessionManager* LSSessionManager::_instance = 0;
u_int32_t LSSessionManager::MinSeqNumber = 1;
u_int32_t LSSessionManager::MaxSeqNumber = UINT_MAX;

extern "C" {
    void sigAlrm(int sig)
    {

    }
    void LSSessionManager_destroy(void)
    {
        LSSessionManager::destroy();
    }
};

LSSessionManager&
LSSessionManager::instance(u_int16_t itad, TRIPTransMode tMode)
{
    if (!_instance)
    {
        _instance = new LSSessionManager(itad, tMode);
        _instance->init();
    }
    return (*_instance);
}

void 
LSSessionManager::destroy(void)
{
    //Clean-up rest of the Singeltons also
    DecisionAgent::destroy();
    OutgoingPolicyAgent::destroy();
    DecisionProxyAgent::destroy();
    TRIBAgent::destroy();

    delete LSSessionManager::_instance;
    LSSessionManager::_instance = 0;
}

LSSessionManager::LSSessionManager(u_int16_t itad, TRIPTransMode tMode)
{
    _transMode = tMode;
    char buf[256];
    gethostname(buf, 256);
    NetworkAddress na(buf);
    _tripItadKey = TRIPItadKey(itad, TRIPId(na.getIp4Address ())); 
    _localIp = na.getIpName();

    if (atexit(LSSessionManager_destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    };

    _topology = new LSTopology(getTripId());

    _serverSocket = new TcpServerSocket(LS_PORT);

    signal(SIGALRM, sigAlrm);

    _seqNum = LSSessionManager::MinSeqNumber;
    _transactionId = 0;

    ///Start the LSApiServer on port 8000, that is the default
    ///port.Can come from provisioning - TODO
    LSApiServer::instance(LS_API_DEF_PORT);
  
    //Register for callBack 
    IncomingPolicyAgent::instance().registerForFlooding(LSSessionManager::floodRoutesCB);
    DecisionAgent::instance().registerForFlooding(LSSessionManager::floodRoutesCB);
}

LSSessionManager::~LSSessionManager()
{
    if(_thread.getPtr() != 0)
    {
        _thread->exit();
    }
    //_serverSocket->close();
}

void
LSSessionManager::run()
{
    ///If threadRun is already running, call to run
    //should be discarded
    if(_thread.getPtr() == 0)
    {
        processMsg();
    }
    else
    {
        cpLog(LOG_ALERT, "Already processing messages in separate thread");
        assert(0);
    }
}

void
LSSessionManager::threadRun()
{
    _thread = new VThread();
    _thread->spawn(LSSessionManager::threadRunImpl, this);
}

void*
LSSessionManager::threadRunImpl(void* args)
{
    LSSessionManager* self = static_cast<LSSessionManager*>(args);
    self->processMsg();
    return (NULL);
}

void
LSSessionManager::processMsg()
{
    for (; ; )
    {
        cpLog(LOG_DEBUG, "Accepting connection");
        Connection conn;
        _serverSocket->accept(conn);
        try
        {
             if(!_lsLocator.isTrustedLs(conn.getIp()))
             {
                 cpLog(LOG_DEBUG, "LS is not trusted, ignoring the connection");
                 conn.close();
                 continue;
             }
             processOpenRequest(conn);
        }
        catch(TRIPBadDataException& e)
        {
            cpLog(LOG_ALERT, "Bad open message, reason:%s",
                             e.getDescription().c_str());
        }
        catch(VException& e)
        {
            cpLog(LOG_ALERT, "Caught exception, reason:%s",
                             e.getDescription().c_str());
        }
    }
}

void
LSSessionManager::processOpenRequest(Connection& conn) throw (VNetworkException&, TRIPBadDataException&)
{
    string destIp = conn.getIp();
    cpLog(LOG_DEBUG, "processOpenRequest from IP:%s", destIp.c_str());
    Sptr<LSPeerSession> pSession = locateExistingPeerSession(destIp);
    if(pSession.getPtr() != 0)
    { 
        cpLog(LOG_DEBUG, "Found existing PeerSession for (%s)", destIp.c_str());
        pSession->handleIncomingConnection(conn);
    }
    else
    {
        cpLog(LOG_DEBUG, "Creating new PeerSession for (%s)", destIp.c_str());
        NetworkAddress addr(conn.getIp(), LS_PORT);
        pSession = createSession(addr);
        lockSessionList();
        _sessionList.push_back(pSession);
        unlockSessionList();
        pSession->handleIncomingConnection(conn);
        pSession->start();
    }
    ostrstream oStr;
    oStr << *pSession << ends;
    cpLog(LOG_DEBUG, "Session initiation (%s)", oStr.str());
    oStr.freeze(false);
}

Sptr<LSPeerSession>
LSSessionManager::createSession(const NetworkAddress& addr)
{
    Sptr<LSPeerSession> retSession=0;
    switch(_transMode)
    {
        case TRIP_TRANS_SND:
            retSession  = new LSSendOnlySession(_tripItadKey.getItad(), _tripItadKey.getTripId(), addr);
        break;
        case TRIP_TRANS_RCV:
            retSession  = new LSReceiveOnlySession(_tripItadKey.getItad(), _tripItadKey.getTripId(), addr);
        break;
        default:
            retSession  = new LSSendReceiveSession(_tripItadKey.getItad(), _tripItadKey.getTripId(), addr);
        break;
    }
    return retSession;
}

void 
LSSessionManager::init()
{
    //initialize the capability list, list of protocol
    //supported by the LS
    //TODO- this list is currently hard-coded, but in future 
    //should come from Provisioning, Also the address family supported
    //Should also come from Provisioning
    _protoCapabilityList.push_back(TRIP_APP_PROTO_SIP);

    _addrFamilyList.push_back(TRIP_ADDR_POTS);
    _addrFamilyList.push_back(TRIP_ADDR_ROUTING_NUM);

    ///Find out all LSs out there and intiate a session
    //TODO
    const LSLocator::LSList& lsList = _lsLocator.getLSList();
    for(LSLocator::LSList::const_iterator itr = lsList.begin();
                itr != lsList.end(); itr++)
    {
        NetworkAddress addr((*itr));
        Sptr<LSPeerSession> sSession = createSession(addr);
        lockSessionList();
        _sessionList.push_back(sSession);
        unlockSessionList();
        sSession->start();
    }
}

ostream&
operator<<(ostream& os, const LSSessionManager& sessionManager)
{
    os << "LS Session list:" << endl;
    sessionManager.lockSessionList();
    for(LSPeerList::const_iterator itr = sessionManager._sessionList.begin(); 
              itr != sessionManager._sessionList.end(); itr++)
    {
        os << *(*itr) << endl;
    }
    sessionManager.unlockSessionList();
    return os;
}

Sptr<LSPeerSession> 
LSSessionManager::locateExistingPeerSession(const string& destIp)
{
    cpLog(LOG_DEBUG, "LSSessionManager::locateExistingPeerSession");
    Sptr<LSPeerSession> retVal;
    lockSessionList();
    for(LSPeerList::const_iterator itr = _sessionList.begin(); 
              itr != _sessionList.end(); itr++)
    {
        if((*itr)->peerExists(destIp))
        {
            retVal = (*itr);
            break;
        }
    }
    unlockSessionList();
    return retVal;
}

bool 
LSSessionManager::locateExistingPeerSession(u_int16_t itad, const TRIPId& tripId)
{
    bool retVal = false;
    lockSessionList();
    for(LSPeerList::const_iterator itr = _sessionList.begin(); 
              itr != _sessionList.end(); itr++)
    {
        if((*itr)->peerExists(itad, tripId))
        {
            retVal = true;
            break;
        }
    }
    unlockSessionList();
    return retVal;
}

bool
LSSessionManager::externalSessionExists() const
{
    bool retVal = false;
    lockSessionList();
    for(LSPeerList::const_iterator itr = _sessionList.begin();
              itr != _sessionList.end(); itr++)
    {
        if ((*itr)->isEstablished() && (!((*itr)->isPeerInternal())))
        {
            retVal = true;
            break;
        }
    }
    unlockSessionList();
    return retVal;
}


void 
LSSessionManager::floodRoutesCB(Sptr <TRIPWorkSet> workSet,
                    TRIPCallbackReason reason, const TRIPItadKey& changedSrc)
{
    LSSessionManager::instance().floodRoutes(workSet, reason, changedSrc);
}

void 
LSSessionManager::floodRoutes(Sptr <TRIPWorkSet> workSet, 
                      TRIPCallbackReason reason, const TRIPItadKey& changedSrc)
{
    bool isExternal = false;
    if(reason == TRIP_CB_EXT_TRIB_UPDATE)
    {
        isExternal = true;
        cpLog(LOG_DEBUG, "Flooding routes as (Ext-trib) changed");
    }
    else
    {
        cpLog(LOG_DEBUG, "Flooding routes as (%s) changed", 
                          changedSrc.getDescription().c_str());
    }

    if (!(workSet.getPtr()))
    {
        assert(0);
    }
    const RouteSetMap& workSetMap = workSet->getRouteSet();
    if (workSetMap.size() == 0)
    {
        return;
    }
    try
    {
        TRIPRouteSet routeSet;
        for (RouteSetMap::const_iterator wItr = workSetMap.begin();
                                wItr != workSetMap.end(); wItr++)
        {
            routeSet = (*wItr).second;
            Sptr <UpdateMsgList> msgList = getFloodMsgList(routeSet, false,
                                                    isExternal);
            assert(msgList.getPtr());
        
            for( LSPeerList::iterator itr = _sessionList.begin();
                              itr != _sessionList.end(); itr++)
            {
                Sptr<LSPeerSession> session = (*itr);

                //If session is not in the establised state, ignore
                if(!session->isEstablished()) continue;

                ///Only send update to peers that has not caused the flooding
                ///If external then send to all internal peers
                if(session->isPeerInternal())
                {
                    session->sendUpdateMsgList(msgList);
                }
            }
        }
    }
    catch(VMissingDataException& e)
    {
        cpLog(LOG_DEBUG, "(%s)", e.getDescription().c_str());
    }
}

Sptr <UpdateMsgList>
LSSessionManager::getFloodMsgList(const TRIPRouteSet& floodSet, 
                                        bool initialFlood, bool isExternal)
{
    if (floodSet.size() == 0)
    {
        return 0;
    }
    u_int8_t flg = 0;
    flg |= TRIP_FLG_LINK_STATE;
    TRIPId id;
    TRIPId localTripId = LSSessionManager::instance().getTripId();
    TRIPItadKey itadKey = LSSessionManager::instance().getTripItadKey();
    u_int32_t seq = 0;
    u_int32_t prevTranId = 0;
    bool msgToSend = false;
    bool withdrawn = false;
    Sptr <UpdateMsgList> retList = new UpdateMsgList();
    Sptr<TRIPUpdateMsg> uMsg;
    Sptr<TRIPWithdrawnRoute> wRoute;
    Sptr<TRIPReachableRoute> rRoute;
    Sptr<TRIPRouteElement> route;
    if ( cpLogGetPriority() >= LOG_DEBUG )
    {
        for(TRIPRouteSet::iterator iitr = floodSet.begin();
                                                        iitr != floodSet.end(); iitr++)
        {
            (*iitr)->prettyDump(cerr);
        }
    }
    for(TRIPRouteSet::iterator itr = floodSet.begin();
                                    itr != floodSet.end(); itr++)
    {
        withdrawn = false;
        if ((*itr)->isWithdrawn())
        {
            withdrawn = true;
        }
        if (initialFlood)
        {
            if (withdrawn) continue;
        }
        if (((*itr)->getTransactionId() == prevTranId) && (msgToSend))
        {
            TRIPRoutes tRoute((*itr)->getAddrFamily(), (*itr)->getAppProtocol(),                               TRIPData((*itr)->getPrefix()));
            if (withdrawn)
            {
                *wRoute << tRoute;
            }
            else
            {
                *rRoute << tRoute;
            }
        }
        else
        {
            prevTranId = (*itr)->getTransactionId();
            if (msgToSend)
            {
                uMsg = new TRIPUpdateMsg(wRoute, rRoute, route->getAttributes());
                retList->push_back(uMsg);
            }

            if (initialFlood)
            {
                id = (*itr)->getOriginator().getTripId();
                seq = (*itr)->getSeqNumber();
            }
            else
            {
                if (isExternal)
                {
                    id = localTripId;
                }
                else
                {
                    id = (*itr)->getOriginator().getTripId();
                }
                seq = getNextSeqNumber(isExternal, (*itr));
            }
            wRoute = new TRIPWithdrawnRoute(flg, id, seq);
            rRoute = new TRIPReachableRoute(flg, id, seq);
            route = (*itr);
            msgToSend = true;
            TRIPRoutes tRoute((*itr)->getAddrFamily(), (*itr)->getAppProtocol(),                               TRIPData((*itr)->getPrefix()));
            if (withdrawn)
            {
                *wRoute << tRoute;
            }
            else
            {
                *rRoute << tRoute;
            }
        }
        if (!(initialFlood))
        {
            (*itr)->unsetChanged(T_FLOOD_BIT);
            (*itr)->setSeqNumber(seq);
            if(isExternal) (*itr)->setOriginator(itadKey);
        }
    }
    if (route.getPtr())
    {
        uMsg = new TRIPUpdateMsg(wRoute, rRoute, route->getAttributes());
        retList->push_back(uMsg);
    }
    cpLog(LOG_DEBUG, "getFloodMsgList - Number of flood msgs is (%d) ",
                          retList->size());
    return retList;
}


Sptr<TRIPUpdateMsg>
LSSessionManager::getNextMsgToFlood(const TRIPTrieCache& cache, bool startFlooding)
{
    //TODO may need to remove
    Sptr<TRIPUpdateMsg> retUMsg = 0;
    Sptr<TRIPRouteElement> element = 0;
    Sptr<TRIPRouteSet> wSet = 0;
    Sptr<TRIPRouteSet> rSet = 0;

    TRIPId id = LSSessionManager::instance().getTripId();
    TRIPItadKey itadKey = LSSessionManager::instance().getTripItadKey();
    u_int8_t flg =  TRIP_FLG_LINK_STATE;
    u_int32_t seqNum = 0;

    retUMsg = new TRIPUpdateMsg;
    TRIPProtoKey key(TRIP_APP_PROTO_SIP, TRIP_ADDR_POTS);
    wSet = cache.getWithdrawnRoutesChangedNeedsFlooding(key); 
    if((wSet.getPtr() != 0) && wSet->size())
    {
        //Withdrawn routes found,get the first route element and query
        //for reachable routes for the same transactionId
        element = *(wSet->begin());
        u_int32_t trId = element->getTransactionId();
        seqNum = getNextSeqNumber(startFlooding, element);
        rSet = cache.getReachableRoutesChangedNeedsFlooding(trId, key);
    }
    else
    {
        //No Withdrawn routes found,however check if there are any reachable
        //routes that needs to be put in the UPDATE message
        rSet = cache.getReachableRoutesChangedNeedsFlooding(key) ;
        if((rSet.getPtr() != 0) && rSet->size())
        {
            element = *(rSet->begin());
            seqNum = getNextSeqNumber(startFlooding, element);
        }
    }
    Sptr<TRIPWithdrawnRoute> wRoute = new TRIPWithdrawnRoute(flg, id, seqNum);
    Sptr<TRIPReachableRoute> rRoute = new TRIPReachableRoute(flg, id, seqNum);
    processRoutes(wRoute, wSet, itadKey, seqNum, startFlooding); 
    processRoutes(rRoute, rSet, itadKey, seqNum, startFlooding);
    retUMsg->addWithdrawnRoute(wRoute);
    retUMsg->addReachableRoute(rRoute);

    //If we have created the UPDATE message add the rest of the attribute to the
    //update message
    assert(element.getPtr());
    const AttributeSet& aSet = element->getAttributes();
    for(AttributeSet::const_iterator itr = aSet.begin();
             itr != aSet.end(); itr++)
    {
        retUMsg->addAttribute((*itr));
    }
    return retUMsg;
}

void
LSSessionManager::processRoutes(Sptr<TRIPRouteContainer> routeContainer,
                                         Sptr<TRIPRouteSet> rSet, 
                                         const TRIPItadKey& itadKey, 
                                         u_int32_t seqNum, bool startFlooding)
{
    if((rSet.getPtr() == 0 ) || (rSet->size() == 0) ) return;

    //Iterate throgh each route element and make it part of the withdrawn route
    for(TRIPRouteSet::iterator itr = rSet->begin(); 
                               itr != rSet->end(); itr++)
    {
        TRIPRoutes tRoute((*itr)->getAddrFamily(), (*itr)->getAppProtocol(),
                       TRIPData((*itr)->getPrefix()));
        *routeContainer << tRoute;
        (*itr)->unsetChanged(T_FLOOD_BIT);
        (*itr)->setSeqNumber(seqNum);
        if(startFlooding) (*itr)->setOriginator(itadKey);
    }
}

u_int32_t
LSSessionManager::getNextSeqNumber(bool startFlooding, Sptr<TRIPRouteElement> data)
{
    //If flooding is being startFlooding 
    //If the Originator of given route element is the same same this
    //LS, increment the seqNumber.
    u_int32_t seqNum;
    if(startFlooding)
    {
        //Originating a new flooding, get a new sequence number
        seqNum = LSSessionManager::instance().getSeqNumber();
    }
    else
    {
        ///If it is previous originator of the route then increment the
        //seq. number 
        seqNum = data->getSeqNumber();
        if(data->getOriginator() == getTripItadKey())
        {
            seqNum++;
        }
    }
    return seqNum;
}

void
LSSessionManager::sendTopologyUpdateOmit(const TRIPItadKey& itadKey, 
                                         Sptr <TRIPUpdateMsg> uMsg ) 
{
    if (uMsg.getPtr())
    {
        lockSessionList();
        for(LSPeerList::const_iterator itr = _sessionList.begin(); 
              itr != _sessionList.end(); itr++)
        {
            if(itadKey == (*itr)->getPeerItadKey()) continue;
            if((*itr)->isEstablished() && (*itr)->isPeerInternal())
            {
                (*itr)->sendUpdate(*uMsg);
            }
        }
        unlockSessionList();
    }
}

void
LSSessionManager::clearAdjTribIn(const TRIPItadKey& itadKey)
{
    cpLog(LOG_DEBUG, "LSSessionManager::clearAdjTribIn");
    //If Peer is an internal peer,
    //Need to send topology updates to all internal peers
    if(itadKey.isInternal())
    {
        //Remove it from the topology set
        LSConnectionList removeList;
        _topology->removeConnection(itadKey.getTripId(), removeList);
        //Dump the _topology
        ostrstream oStr;
        _topology->dump(oStr);
        oStr << ends;
        cpLog(LOG_INFO, "TOPOLOGY: %s", oStr.str());
        oStr.freeze(false);

        if(removeList.size())
        {
            //Remove the adj-trib-ins
            for(LSConnectionList::iterator itr = removeList.begin();
                                       itr != removeList.end(); itr++)
            {
                TRIPItadKey key(getItad(), (*itr));
                cpLog(LOG_DEBUG, "Destroying Adj-trib-in for (%s)",
                              key.getDescription().c_str());
                TRIBAgent::instance().destroyAdjTribIn(key);
            }
        }
        Sptr <LSConnectionList> cList = _topology-> 
                                                getConnectionList(getTripId());
        if (cList.getPtr())
        {
            Sptr<TRIPUpdateMsg> uMsg = createTopologyUpdateMsg();
            sendTopologyUpdateOmit(itadKey, uMsg);
        }
    }
    else
    {
        Sptr<TRIPUpdateEvent> event = new TRIPUpdateEvent(0, itadKey, 
                                                        TRIP_CONNECTION_LOST);
        DecisionProxyAgent::instance().addToQueue(event);
        TRIBAgent::instance().destroyAdjTribOut(itadKey);
    }
}

Sptr<TRIPUpdateMsg>
LSSessionManager::createTopologyUpdateMsg(Sptr<TRIPItadTopology> topologyAttr)
{
    Sptr<TRIPUpdateMsg> uMsg = 0;
    Sptr<LSConnectionList> cList;
    if (!(topologyAttr.getPtr()))
    {
        cList = _topology->getConnectionList(getTripId());
        if (!(cList.getPtr()))
        {
            return uMsg;
        }
    }
    //Create an Update message with Topology attribute to send to
    //rest of the internal peer sessions.
    uMsg = new TRIPUpdateMsg();
    TRIPId id = getTripId();
    Sptr<TRIPWithdrawnRoute> wRoute =
               new TRIPWithdrawnRoute(TRIP_FLG_LINK_STATE, id, 0);
    Sptr<TRIPReachableRoute> rRoute =
               new TRIPReachableRoute(TRIP_FLG_LINK_STATE, id, 0);
    uMsg->addWithdrawnRoute(wRoute);
    uMsg->addReachableRoute(rRoute);

    //Add all the mendatory attributes

    Sptr<TRIPNextHopServer> nServer =
                        new TRIPNextHopServer(getItad(),TRIPData(""));
    uMsg->addAttribute(nServer);
    Sptr<TRIPAdvertisePath> advPath = new TRIPAdvertisePath();
    Sptr<TRIPRoutedPath> routePath = new TRIPRoutedPath();
    uMsg->addAttribute(advPath);
    uMsg->addAttribute(routePath);

    if(topologyAttr != 0)
    {
        uMsg->addAttribute(topologyAttr);
    }
    else
    {
        u_int32_t seqNum = getNextSeqNumber(true, 0);
        Sptr<TRIPItadTopology> itadTopology =
               new TRIPItadTopology(id, seqNum, TRIP_FLG_LINK_STATE);
        //Get the list of TRIPIds and form the ItadTopology Attribute
        for(LSConnectionList::iterator itr = cList->begin();
                 itr != cList->end(); itr++)
        {
            itadTopology->addTripId((*itr));
        }
        uMsg->addAttribute(itadTopology);
    }

    ostrstream ostr;
    ostr << "Topology Update Msg:" << endl;
    ostr << *uMsg << endl;
    ostr << ends;
    cpLog(LOG_DEBUG, "%s", ostr.str());
    ostr.freeze(false);
    return (uMsg);
}

void
LSSessionManager::handleTopologyChanges(const TRIPItadKey& srcItadKey,
                      Sptr<TRIPItadTopology> itadTopology)
{
    cpLog(LOG_DEBUG, "LSSessionManager::handleTopologyChanges");
    TRIPItadTopologySet recSet;
    LSConnectionList addList, removeList;
    bool flood = _topology->updateTopology(itadTopology, addList, removeList);
    if(removeList.size())
    {
        //Remove the adj-trib-ins
        for(LSConnectionList::iterator itr = removeList.begin(); 
                                       itr != removeList.end(); itr++)
        {
            TRIPItadKey key(getItad(), (*itr));
            cpLog(LOG_DEBUG, "Destroying Adj-trib-in for (%s)",
                              key.getDescription().c_str());
            TRIBAgent::instance().destroyAdjTribIn(key);
        }
    }
    if(flood)
    {
        Sptr<TRIPUpdateMsg> uMsg = createTopologyUpdateMsg(itadTopology);
        //Since we need to send to every internal peer session, 
        TRIPItadKey none;
        sendTopologyUpdateOmit(none, uMsg);
    }

    ostrstream oStr;
    _topology->dump(oStr);
    oStr << ends;
    cpLog(LOG_INFO, "TOPOLOGY: %s", oStr.str());
    oStr.freeze(false);
}

void 
LSSessionManager::addTopology(const TRIPItadKey& lKey, const TRIPItadKey& peerKey)
{
    cpLog(LOG_DEBUG, "LSSessionManager::addTopology");
    _topology->addConnection(lKey.getTripId(), peerKey.getTripId());
    Sptr<TRIPUpdateMsg> uMsg = createTopologyUpdateMsg(); 
    TRIPItadKey none;
    sendTopologyUpdateOmit(none, uMsg);                                    
 
    ostrstream oStr;
    _topology->dump(oStr);
    oStr << ends;
    cpLog(LOG_INFO, "TOPOLOGY: %s", oStr.str());
    oStr.freeze(false);
}

