#ifndef LSSessionManager_hxx
#define LSSessionManager_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LSSessionManager_hxx_Version = 
"$Id: LSSessionManager.hxx,v 1.27 2001/03/06 03:02:31 kevinmc Exp $";

#include "global.h"
#include <vector>
#include <string>

#include "vtypes.h"
#include "Sptr.hxx"
#include "TRIPItadKey.hxx"
#include "Connection.hxx"
#include "TRIPBadDataException.hxx"
#include "TRIPTransCapability.hxx"
#include "LSLocator.hxx"
#include "TRIPDef.hxx"
#include "LSTopology.hxx"
#include "VMutex.h"

class LSPeerSession;
class NetworkAddress;
class TRIPUpdateMsg;
class TRIPRouteContainer;

/** 
    Singelton class,implements the  connection management at the LS.It is 
    listening on port 6069 for any incoming open request.When a request 
    arrives, parses it find the capabilities.Based on the capabilities 
    creates a LSPeerSession object and delegates the new connection to the 
    LSPeerSession.LSSessionManager can also initates an Open connection to its
    peers, in that case based on the capability it can support, creates a
    LSPeerSession and delegates the session process to PeerSession.
    Session managers runs either as a separate thread in 
    proxy or can run as standalone application.
*/
class LSSessionManager {
    public:
        ///
        static LSSessionManager& instance(u_int16_t itad=0, TRIPTransMode tMode = TRIP_TRANS_SND_RCV);

        ///
        string className() { return "LSSessionManager"; }

        ///Destructor
        virtual ~LSSessionManager();

        /**
           Frees the memory associated with singelton instance.
           gets register to atexit() function at the time of creation.
        */
        static void destroy(void);

        ///
        friend ostream& operator<<(ostream& os, const LSSessionManager& lslocator);

        ///Returns true if session with given ITAD and tripId exists
        bool locateExistingPeerSession(u_int16_t itad, const TRIPId& tripId);

        /**Returns pointer to the PeerSession ,
         * if a PeerSession exists for a given destination.
         */
        Sptr<LSPeerSession>  locateExistingPeerSession(const string& destIp);

        ///Indicates if any session is established with an external Peer
        bool externalSessionExists() const;

        ///
        const TRIPItadKey& getTripItadKey() const { return _tripItadKey; };

        ///
        u_int16_t getItad() const { return _tripItadKey.getItad(); };

        ///
        const TRIPId& getTripId() const { return _tripItadKey.getTripId(); };

        ///
        TRIPTransMode getTransMode() const { return _transMode; };

        ///
        const string& getLocalIp() const { return _localIp; }

        /**Runs in while loop accepting OPEN connections in forground in the 
         * main thread.Should be used ONLY when running LS
         * as stand-alone application.
         */
        void run();

        /**Creates a separate thread to run the while loop accepting 
           connections.The method forces LS to run in its
           own thread.Should ONLY be used when running LS as part of any
           other application, for example within a Proxy.
         */
        void threadRun();

        ///
        const LSPeerList& getSessionList() const { return _sessionList; };

        ///
        void lockSessionList() const { _sessionListMutex.lock(); };

        ///
        void unlockSessionList()  const { _sessionListMutex.unlock(); };

        ///
        static u_int32_t MinSeqNumber;

        ///
        static u_int32_t MaxSeqNumber;

        ///
        inline u_int32_t getSeqNumber() { return ++_seqNum; };

        ///Gets called when the ADJ-TRIB-IN for tribIn changed.
        void floodRoutes(Sptr <TRIPWorkSet> workSet, TRIPCallbackReason reason,
                                                    const TRIPItadKey& key);

        ///
        static void floodRoutesCB(Sptr <TRIPWorkSet> workSet,
                      TRIPCallbackReason reason, const TRIPItadKey& changedSrc);

        ///
        inline u_int32_t getNextTransactionId() 
        { 
            return ++_transactionId; 
        };

        Sptr <UpdateMsgList> getFloodMsgList(const TRIPRouteSet& floodSet,
                                        bool initialFlood, bool isInternal);

        ///
        u_int32_t getNextSeqNumber(bool startFlooding, Sptr<TRIPRouteElement> data);

        ///
        const LSLocator& getLsLocator() const { return _lsLocator; };

        ///
        const ProtoCapabilityList& getProtoCapabilityList() const 
        { 
            return _protoCapabilityList; 
        };

        ///
        const AddrFamilyList& getAddrFamilyList() const 
        { 
            return _addrFamilyList; 
        };

        ///
        void sendTopologyUpdateOmit(const TRIPItadKey& srcItadKey, 
                                    Sptr <TRIPUpdateMsg> uMsg );

        ///
        void handleTopologyChanges(const TRIPItadKey& srcItadKey,
                      Sptr<TRIPItadTopology> itadTopology);

        ///
        void clearAdjTribIn(const TRIPItadKey& itadKey);

        ///
        Sptr<TRIPUpdateMsg> createTopologyUpdateMsg(Sptr<TRIPItadTopology> topo=0);

        ///
        void addTopology(const TRIPItadKey& lKey, const TRIPItadKey& peerKey);
    protected:

        ///
        Sptr<TRIPUpdateMsg> getNextMsgToFlood(const TRIPTrieCache& cache, 
                                              bool startFlooding);

        ///
        void processRoutes(Sptr<TRIPRouteContainer> routeContainer,
                                         Sptr<TRIPRouteSet> rSet,
                                         const TRIPItadKey& itadKey,
                                         u_int32_t seqNum, bool startFlooding);


        ///
        Sptr<LSPeerSession> createSession(const NetworkAddress& addr);

        ///
        static void* threadRunImpl(void* args);

        ///
        void processMsg();

        ///
        void processOpenRequest(Connection& conn) throw (VNetworkException&, TRIPBadDataException&);

        ///
        void init();


         ///
         LSSessionManager(u_int16_t itad, TRIPTransMode tMode);


         ///
         static LSSessionManager* _instance;

         ///
         Sptr<TcpServerSocket>  _serverSocket;

         ///
         Sptr<VThread>          _thread;

         ///
         LSPeerList _sessionList;

         ///
         mutable VMutex     _sessionListMutex;

         ///
         TRIPItadKey  _tripItadKey;

         ///
         u_int32_t _seqNum;

         ///
         u_int32_t  _transactionId;

         ///
         LSLocator  _lsLocator;

         ///
         ProtoCapabilityList    _protoCapabilityList;

         ///
         AddrFamilyList        _addrFamilyList;

         ///
         TRIPTransMode _transMode;

         ///
         Sptr<LSTopology>      _topology;

         ///
         string                _localIp;

};

#endif
