/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const OutgoingPolicyAgent_cxx_Version = 
"$Id: OutgoingPolicyAgent.cxx,v 1.23 2001/03/07 22:03:04 kevinmc Exp $";

#include <iostream.h>
#include <strstream.h>
#include "LSSessionManager.hxx"
#include "LSPeerSession.hxx"
#include "OutgoingPolicyAgent.hxx"
#include "OutgoingPolicyTimerAgent.hxx"
#include "DecisionProxyAgent.hxx"
#include "TRIBAgent.hxx"
#include "Aggregator.hxx"
#include "Advertiser.hxx"
#include "TRIPUpdateMsg.hxx"
#include "TRIPItadKey.hxx"
#include "TRIPRouteElement.hxx"
#include "TRIPDef.hxx"
#include <vector>
#include "cpLog.h"

OutgoingPolicyAgent* OutgoingPolicyAgent::_instance = 0;

OutgoingPolicyAgent&
OutgoingPolicyAgent::instance()
{
    if (!_instance)
    {
        _instance = new OutgoingPolicyAgent();
    }
    return (*_instance);
}

OutgoingPolicyAgent::OutgoingPolicyAgent()
    : _shutdown(false)
{
    _thread = new VThread();
    _thread->spawn(OutgoingPolicyAgent::run, this);
}


OutgoingPolicyAgent::~OutgoingPolicyAgent()
{
    _shutdown = true;
    runPolicy->Signal();

    if (_thread.getPtr() != 0)
    {
        _thread->exit();
    }
}

void
OutgoingPolicyAgent::destroy(void)
{
    delete OutgoingPolicyAgent::_instance;
    OutgoingPolicyAgent::_instance = 0;
}

void*
OutgoingPolicyAgent::run(void* args)
{
    OutgoingPolicyAgent* self = (OutgoingPolicyAgent*)args;
    self->init();
    while (true)
    {
        self->runPolicy->Wait(); // wait for OutgoingPolicyTimerAgent
        if(self->_shutdown) break;
        self->policyEnforcer();
        self->timerAgt->startTimer(); // restart Timer after advertisement
    }
    return (0);
}
 
void
OutgoingPolicyAgent::addNewPeerKey(const TRIPItadKey& key)
{
    newPeerKeyList.push_back(key);
}

void 
OutgoingPolicyAgent::policyEnforcer()
{
    cpLog(LOG_DEBUG, "OutgoingPolicyAgent::process");
    LSSessionManager& sessionMgr = LSSessionManager::instance();
    if (!(sessionMgr.externalSessionExists()))
    {
        cpLog(LOG_DEBUG, "OutgoingPolicyAgent - no external Peers exist");
        return;
    }
    TRIBAgent& tribAgent = TRIBAgent::instance();
    Sptr <TRIPTrieCache> locTrib = tribAgent.getLocTrib();
    if (locTrib->isUpdated()) // locTrib updated, run OutgoingPolicy
    {
        Sptr <Aggregator> aggAgt = new Aggregator();
        locTrib->writeLock();
        locTrib->setUpdated(false);
        aggAgt->process(*locTrib);
        locTrib->unlock();
    }

    cpLog(LOG_DEBUG, "OutgoingPolicyAgent - aggregation complete");
    processNewPeers();
    const ProtoCapabilityList& protoList = sessionMgr.getProtoCapabilityList();
    const AddrFamilyList& addrFamilyList = sessionMgr.getAddrFamilyList();
    const LSPeerList& sessionList = sessionMgr.getSessionList();
    Sptr <TRIPTrieCache> adjTribOut;

    for (AddrFamilyList::const_iterator addrItr = addrFamilyList.begin();
                                    addrItr != addrFamilyList.end(); addrItr++)
    {
        for (ProtoCapabilityList::const_iterator itr = protoList.begin();
                                            itr != protoList.end(); itr++)
        {
            TRIPProtoKey pKey((*itr), (*addrItr));
         
            bool moreRoutesToProcess = false;

            do
            {
                locTrib->readLock();
                Sptr <TRIPRouteSet> src = locTrib->getNextOutRouteSet
                                                    (moreRoutesToProcess, pKey);
                sessionMgr.lockSessionList();

                for(LSPeerList::const_iterator itr = sessionList.begin(); 
                                            itr != sessionList.end(); itr++)
                {
                    const TRIPItadKey& sessionItad = (*itr)->getPeerItadKey();
                    if ((*itr)->isEstablished() && (!((*itr)->isPeerInternal())))
                    {
                        try
                        {
                            adjTribOut = tribAgent.getTribOut(sessionItad);
                        }
                        catch (VMissingDataException e)
                        {
                            cpLog(LOG_INFO, "OutgoingPolicyAgent - no adj-TribOut exists for (%s)",
                                       sessionItad.getDescription().c_str());
                            continue;
                        }
                        adjTribOut->writeLock();
                        for(TRIPRouteSet::iterator srcItr = src->begin();
                                                srcItr != src->end(); srcItr++)
                        {
                            processPolicyRules((*srcItr), adjTribOut, 
                                                                sessionItad);
                        }
                        adjTribOut->unlock();
                    }
                }
                sessionMgr.unlockSessionList();
                locTrib->unlock();
            }while (moreRoutesToProcess);
        }
    }
    Sptr <Advertiser> adv = new Advertiser();
    adv->process();
}    


void
OutgoingPolicyAgent::processPolicyRules(Sptr <TRIPRouteElement> route,
                                            Sptr <TRIPTrieCache> dest,
                                            const TRIPItadKey& itad)
{
    // if route has been received from itad then do not return the route
    if (route->getSource() == itad)
    {
        //if selected route is from itad make sure that previous is Withdrawn
        TRIPRouteSet retSet;
        if (dest->matchDestination(route, retSet))
        {
            (*(retSet.begin()))->setWithdrawn();
        }
        return;
    }
    Sptr<TRIPRouteElement> newRoute = new TRIPRouteElement(*route);
    newRoute->unsetChanged(RouteChangeBitMask(T_ALL_BIT));
    newRoute->setChanged(RouteChangeBitMask(T_UPDATE_BIT));
    dest->insert(newRoute);
}


void
OutgoingPolicyAgent::processNewPeers()
{
    if (newPeerKeyList.size() > 0)
    {
        for (vector<TRIPItadKey>::const_iterator itr = newPeerKeyList.begin();
                    itr != newPeerKeyList.end(); itr++)
        {
            // check new Peer is valid
            if (LSSessionManager::instance().locateExistingPeerSession
                                        ((*itr).getItad(), (*itr).getTripId()))
            {
                populateAdjTribOut((*itr));
            }
        }
        newPeerKeyList.clear();
    }
} 

void
OutgoingPolicyAgent::populateAdjTribOut(const TRIPItadKey& key)
{
    cpLog(LOG_DEBUG, "OutgoingPolicyAgent::populateAdjTribOut for (%s)",
                                                key.getDescription().c_str());
    TRIBAgent& tribAgent = TRIBAgent::instance();
    tribAgent.createAdjTribOut(key, TC_EXTERNAL);
    Sptr <TRIPTrieCache> adjTribOut; 
    try
    {
        adjTribOut = tribAgent.getTribOut(key);
    }
    catch (VMissingDataException e)
    {
        cpLog(LOG_INFO, "OutgoingPolicyAgent - no adj-TribOut exists for (%s)",
                                                key.getDescription().c_str());
        return;
    }
    Sptr <TRIPTrieCache> locTrib = tribAgent.getLocTrib();

    LSSessionManager& sessionMgr = LSSessionManager::instance();
    const ProtoCapabilityList& protoList = sessionMgr.getProtoCapabilityList();
    const AddrFamilyList& addrFamilyList = sessionMgr.getAddrFamilyList();
    locTrib->readLock();
    adjTribOut->writeLock();

    for (AddrFamilyList::const_iterator addrItr = addrFamilyList.begin();
                                    addrItr != addrFamilyList.end(); addrItr++)
    {
        for (ProtoCapabilityList::const_iterator itr = protoList.begin();
                                            itr != protoList.end(); itr++)
        {
            TRIPProtoKey pKey((*itr), (*addrItr));

            Sptr <TRIPRouteSet> aggRouteSet = locTrib->getRouteSet(pKey);
            if (aggRouteSet->size() == 0)
            {
                continue;
            }
    
            for(TRIPRouteSet::iterator itr = aggRouteSet->begin();
                                            itr != aggRouteSet->end(); itr++)
            {
                if ((*itr)->isAggregated()) continue;
                if ((*itr)->isWithdrawn()) continue;
                processPolicyRules((*itr), adjTribOut, key);
            }
        }
    }
    adjTribOut->unlock();
    locTrib->unlock();
}
        
    
    
void
OutgoingPolicyAgent::init()
{
    // TODO need to initialize policy parameters
    runPolicy = new VThreadEvent();
    timerAgt = new OutgoingPolicyTimerAgent(runPolicy);
    timerAgt->setDuration(OUTGOING_POLICY_TIMER);
    timerAgt->startTimer();
}
