/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const TRIBAgent_cxx_Version = 
"$Id: TRIBAgent.cxx,v 1.18 2001/03/22 01:04:28 kevinmc Exp $";


#include <stdio.h>
#include "TRIBAgent.hxx"
#include "cpLog.h"

TRIBAgent* TRIBAgent::_instance = 0;

TRIBAgent&
TRIBAgent::instance()
{
    if (!_instance)
    {
        _instance = new TRIBAgent();
    }
    return (*_instance);
}

void 
TRIBAgent::destroy(void)
{
    delete TRIBAgent::_instance;
    TRIBAgent::_instance = 0;
}

TRIBAgent::TRIBAgent()
{
#ifdef __linux__
    if (atexit(TRIBAgent::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    };
#endif
    _extTrib = new TRIPTrieCache();
    _locTrib = new TRIPTrieCache();
}

TRIBAgent::~TRIBAgent()
{
}

Sptr<TRIPTrieCache> 
TRIBAgent::getTribIn(const TRIPItadKey& key) 
    throw (VMissingDataException&)
{
    if(!_mapOfTribIn.count(key))
    {
        char buf[256];
        sprintf(buf, "Missing ADJ-Trib-In for (%s)", key.getDescription().c_str()); 
        cpLog(LOG_INFO, buf);
        throw VMissingDataException(buf, __FILE__, __LINE__, 0);
    }
    return (_mapOfTribIn[key]);
}

void
TRIBAgent::createAdjTribIn(const TRIPItadKey& key, TCacheType cType)
{
    //First check if TRIB-In already exists for given ItadKey and type,
    //DO nothing if already exists
    //getTribIn() throws exception if TRIB-In is not found.

    Sptr<TRIPTrieCache> tribCache;
    try
    {
        tribCache = getTribIn(key);
    }
    catch (VMissingDataException)
    {
        tribCache = new TRIPTrieCache(key, cType);
        _mapOfTribIn.insert(pair<TRIPItadKey, Sptr<TRIPTrieCache> > (key, tribCache));
    }
}

Sptr<TRIPTrieCache> 
TRIBAgent::getTribOut(const TRIPItadKey& key) 
    throw (VMissingDataException&)
{
    if(!_mapOfTribOut.count(key))
    {
        char buf[256];
        sprintf(buf, "Missing ADJ-Trib-Out for (%s)",key.getDescription().c_str());
        cpLog(LOG_INFO, buf);
        throw VMissingDataException(buf, __FILE__, __LINE__, 0);
    }
    return (_mapOfTribOut[key]);
}

void
TRIBAgent::createAdjTribOut(const TRIPItadKey& key, TCacheType cType)
{
    //First check if TRIB-In already exists for given ItadKey and type,
    //DO nothing if already exists
    //getTribIn() throws exception if TRIB-In is not found.

    Sptr<TRIPTrieCache> tribCache;
    try
    {
        tribCache = getTribOut(key);
    }
    catch (VMissingDataException)
    {
        tribCache = new TRIPTrieCache(key, cType);
        _mapOfTribOut.insert(pair<TRIPItadKey, Sptr<TRIPTrieCache> > (key, tribCache));
    }
}

Sptr <TRIPTrieCache> 
TRIBAgent::getLocTrib() 
{
    return _locTrib;
}

Sptr <TRIPTrieCache> 
TRIBAgent::getExtTrib() 
{
    return _extTrib;
}

Sptr<TRIPTrieList> 
TRIBAgent::getAllInternalTribIns() const
{
    Sptr<TRIPTrieList> ret = new TRIPTrieList;
    for(TRIBTrieCacheIn::const_iterator itr = _mapOfTribIn.begin();
                     itr != _mapOfTribIn.end(); itr++)
    {
        
        if((*itr).second->isInternalRouteCache())
        {
            //omit the internal cache with local key - only used for flooding purposes
            if ((*itr).first.isLocal()) continue;
            ret->push_back((*itr).second);
        }
    }
    return ret;
}

Sptr<TRIPTrieList> 
TRIBAgent::getSelectTrieListOmit(const TRIPItadKey& key)
{
    assert(_mapOfTribIn.count(key));

    Sptr<TRIPTrieList> ret = new TRIPTrieList;

    if((_mapOfTribIn[key]->isExternalRouteCache()) ||
       (_mapOfTribIn[key]->isLocalRouteCache()))
    {
       //Return External + local ommiting the one matching the key
       for(TRIBTrieCacheIn::iterator itr = _mapOfTribIn.begin();
                                 itr != _mapOfTribIn.end(); itr++)
       {
           if(((*itr).first != key) && 
              !((*itr).second->isInternalRouteCache()))
           {
               ret->push_back((*itr).second);
           }
       }
    }
    else if(_mapOfTribIn[key]->isInternalRouteCache())
    {
       //Return ExtTrib + Internal ommiting the one matching the key
       for(TRIBTrieCacheIn::iterator itr = _mapOfTribIn.begin();
                                 itr != _mapOfTribIn.end(); itr++)
       {
           if(((*itr).first != key) && 
              ((*itr).second->isInternalRouteCache()))
           {
               //omit internal cache with local key as it is only used for flooding
               if ((*itr).first.isLocal()) continue;
               ret->push_back((*itr).second);
           }
       }
       ret->push_back(_extTrib);
    }
    return ret;
}

///
void 
TRIBAgent::readLock() const
{
    cpLog(LOG_DEBUG, "Locking all TRIBS read-only");
    _lock.ReadLock();
}

///
void 
TRIBAgent::writeLock() const
{
    cpLog(LOG_DEBUG, "Locking all TRIBS read-write");
    _lock.WriteLock();
}

///
void 
TRIBAgent::unlock() const
{
    cpLog(LOG_DEBUG, "Unlocking all TRIBS");
    _lock.Unlock();
}

void 
TRIBAgent::destroyAdjTribIn(const TRIPItadKey& key)
{
    cpLog(LOG_DEBUG, "TRIBAgent::destroyAdjTribIn");
    writeLock();
    TRIBTrieCacheIn::iterator itr = _mapOfTribIn.find(key);
    if(itr != _mapOfTribIn.end())
    {
        _mapOfTribIn.erase(itr);
    }
    unlock();
}

void
TRIBAgent::destroyAdjTribOut(const TRIPItadKey& key)
{
    cpLog(LOG_DEBUG, "TRIBAgent::destroyAdjTribOut");
    writeLock();
    TRIBTrieCacheIn::iterator itr = _mapOfTribOut.find(key);
    if(itr != _mapOfTribOut.end())
    {
        _mapOfTribOut.erase(itr);
    }
    unlock();
}

void
TRIBAgent::prettyDump(ostream& os) const
{
    cpLog(LOG_DEBUG, "TRIBAgent::prettyDump");
    readLock();
    for(TRIBTrieCacheIn::const_iterator itr = _mapOfTribIn.begin();
                     itr != _mapOfTribIn.end(); itr++)
    {
        string peerType;
        if (((*itr).second)->isExternalRouteCache()) peerType = "External ";
        else if (((*itr).second)->isLocalRouteCache()) peerType = "Local ";
        else if (((*itr).second)->isInternalRouteCache()) peerType = "Internal ";

        os << endl << endl << peerType << "ADJTRIBIN : ITADKEY = " << (*itr).first << endl;
        ((*itr).second)->prettyDump(os);
    }
    unlock();
    os << endl << endl << "EXTTRIB " << endl;
    _extTrib->readLock();
    _extTrib->prettyDump(os);
    _extTrib->unlock();
    os << endl << endl << "LOCTRIB " << endl;
    _locTrib->readLock();
    _locTrib->prettyDump(os);
    _locTrib->unlock();
    readLock();
    for(TRIBTrieCacheIn::const_iterator itr = _mapOfTribOut.begin();
                     itr != _mapOfTribOut.end(); itr++)
    {
        os << endl << endl << "ADJTRIBOUT : ITADKEY = " << (*itr).first << endl;
        ((*itr).second)->prettyDump(os);
    }
    unlock();
}

