/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
#include <stdio.h>
#include <sys/types.h>
#include <regex.h>
#include <time.h>
#include <sys/errno.h>

#include "TRIPRouteElement.hxx"
#include "TRIPUtil.hxx"
#include "LSSessionManager.hxx"
#include "cpLog.h"

TRIPRouteElement::TRIPRouteElement()
    : _timeStamp(time(0)),
      _trId(0),
      _cost(0.0),
      _load(0),
      _seqNum(0),
      _changed(T_UPDATE_BIT),
      _withdrawn(false),
      _itrip(false),
      _aggrStatus(T_NON_AGGREGATED)
{
}

TRIPRouteElement::TRIPRouteElement(const TRIPItadKey& key, 
                  const TRIPRoutes& route, 
                  const Data& nextHopServer,
                  const TRIPUpdateMsg::AttrSet attrSet,
                  bool itrip)
    : _timeStamp(time(0)),
      _trId(0),
      _route(route),
      _nextHop(nextHopServer),
      _cost(0.0),
      _load(0),
      _source(key),
      _seqNum(0),
      _changed(T_UPDATE_BIT),
      _withdrawn(false),
      _itrip(itrip),
      _aggrStatus(T_NON_AGGREGATED)
{
    for(TRIPUpdateMsg::AttrSet::iterator itr = attrSet.begin();
            itr != attrSet.end(); itr++)
    {
        _attrSet.insert(*itr);
    }
}

TRIPRouteElement::TRIPRouteElement(const TRIPItadKey& key,
                  const TRIPRoutes& route,
                  const Data& nextHopServer,
                  bool itrip)
    : _timeStamp(time(0)),
      _trId(0),
      _route(route),
      _nextHop(nextHopServer),
      _cost(0.0),
      _load(0),
      _source(key),
      _seqNum(0),
      _changed(T_UPDATE_BIT),
      _withdrawn(false),
      _itrip(itrip),
      _aggrStatus(T_NON_AGGREGATED)
{
}

TRIPRouteElement::TRIPRouteElement(const TRIPRouteElement& other) 
{
    LockHelper lck(other._mutex);
    copyObj(other);
}

void
TRIPRouteElement::copyObj(const TRIPRouteElement& other)
{
    _timeStamp = time(0);
    _route = other._route;
    _trId = other._trId;
    _nextHop = other._nextHop;
    _cost = other._cost;
    _load = other._load;
    _changed = other._changed;
    _withdrawn = other._withdrawn;
    _seqNum = other._seqNum;
    _source = other._source;
    _itrip = other._itrip;
    _originator = other._originator;
    _aggrStatus = other._aggrStatus;
    _attrSet.clear();
    for(TRIPUpdateMsg::AttrSet::iterator itr = other._attrSet.begin();
            itr != other._attrSet.end(); itr++)
    {
        _attrSet.insert(*itr);
    }
}

TRIPRouteElement& 
TRIPRouteElement::operator=(const TRIPRouteElement& other)
{
    if(this != &other)
    {
        copyObj(other);
    }
    return (*this);
}


ostream& operator<<(ostream& os, const TRIPRouteElement& obj)
{
    char buf[256];
#if defined(__linux__)
    ctime_r(&obj._timeStamp, buf);
#else
    ctime_r(&obj._timeStamp, buf, 255);
#endif
    os.width(25);
    os << "**Route Element**" << endl;
    os.width(25);
    os << "Time created: " << buf ;
    os.width(25);
    os << "Prefix: " << obj._route.getPrefix().c_str() << endl;
    os.width(25);
    os << "Next Hop: " << obj._nextHop.getData() << endl;
    os.width(25);
    os << "Seq number: " << obj._seqNum << endl;
    os.width(25);
    os << "Source : " << obj._source << endl;
    os.width(25);
    os << "Transaction Id : " << obj._trId << endl;
    os.width(25);
    os << "Cost: " << obj._cost << endl;
    os.width(25);
    os << "Load: " << obj._load << endl;
    os.width(25);
    os << "Itrip: " << ((obj._itrip) ? "Yes" : "No") << endl;
    os.width(25);
    os << "Changed: " << ((obj._changed & T_AGGR_BIT) ? "A" : " ");
    os << ((obj._changed & T_UPDATE_BIT) ? "U" : " ");
    os << ((obj._changed & T_PROXY_BIT) ? "P" : " ");
    os << ((obj._changed & T_OUTGOING_BIT) ? "O" : " ");
    os << ((obj._changed & T_FLOOD_BIT) ? "F" : " ");
    os << ((obj._changed & T_EXTTRIB_BIT) ? "E" : " ") << endl;
    os.width(25);
    os << "Withdrawn: " << ((obj._withdrawn) ? "Yes" : "No") << endl;
    os.width(25);
    os << "Aggr Status: " << ((obj._aggrStatus & T_NON_AGGREGATED) ? "N" : " ");
    os << ((obj._aggrStatus & T_AGGREGATED) ? "D" : " ");
    os << ((obj._aggrStatus & T_AGGREGATED_RES) ? "R" : " ") << endl;
    os.width(25);
    os << endl << "**Attributes**" << endl;
    for(TRIPUpdateMsg::AttrSet::iterator itr = obj._attrSet.begin();
            itr != obj._attrSet.end(); itr++)
    {
        os.width(25);
        os << *(*itr) << endl;
    }
    return os;
}

void
TRIPRouteElement::prettyDump(ostream& os) const
{
    _mutex.lock();
    //Print one line summary of the route element
    //in format SOURCE - PROTO - ADDRFamily - NEXTHOP - PREFIX - SEQ - Transaction Id - W/R (withdrawn/reachable)
    os <<  _source ; 
    os << "," << static_cast<int>(getAppProtocol());
    os << "," << static_cast<int>(getAddrFamily());
    os << "," << _nextHop.getData();
    os << "," << _route.getPrefix().c_str();
    os << "," << _seqNum;
    os << "," << _trId;
    //Print the local pref
    const Sptr<TRIPRouteAttr> rAttr = getAttribute(TRIP_LOCAL_PREF);
    Sptr<TRIPLocalPref> lAttr;
    lAttr.dynamicCast(rAttr);
    if(lAttr != 0)
    {
        os << ", Pref:" << lAttr->getPrefNumber();
    }
    _mutex.unlock();
    os << "," << (isWithdrawn() ? "W" : "R") << endl;
}

bool 
TRIPRouteElement::operator==(const TRIPRouteElement& other) const
{
    if ((_source == other._source) &&
        (getPrefix() == other.getPrefix()))
    {
        return true;
    }
    return false;
}

inline bool 
TRIPRouteElement::operator<(const TRIPRouteElement& other) const
{
    //Check if they are from the same ITAD
    //Withing the same ITAD, routes are sorted based on the Prefix

    if (_source < other._source) return true;
    if (_source == other._source)
    {
        if(getPrefix() < other.getPrefix()) return true;
    }
    return false;
}

void
TRIPRouteElement::writeData(int id) const throw(VNetworkException&)
{
    assert(0);
}


const Sptr<TRIPRouteAttr> 
TRIPRouteElement::getAttribute(TRIPAttrType aType) const
{
    Sptr<TRIPRouteAttr> retAttr = 0;
    AttributeSet::iterator itr;
    for(itr = _attrSet.begin(); itr != _attrSet.end(); itr++)
    {
        Sptr<TRIPRouteAttr> attr = (*itr);
        if(attr->getType() == aType)
        {
            return attr;
        }
    } 
    return retAttr;
}

void
TRIPRouteElement::addAttribute(Sptr<TRIPRouteAttr> attr) 
{
    LockHelper lck(_mutex);
    AttributeSet::iterator itr;
    for(itr = _attrSet.begin(); itr != _attrSet.end(); itr++)
    {
        if((*itr)->getType() == attr->getType())
        {
            cpLog(LOG_DEBUG, "Attribute type (%d) already added, ignoring", 
                              attr->getType());
            return;
        }
    }
    _attrSet.insert(attr);
}

bool 
TRIPRouteElement::incrementSeqNum()
{
    //Check if seq number has exceeded the MAXSeqNumber, if so return false
    //so that caller would know it and would like to use info to do any further
    //processing
    if(_seqNum == LSSessionManager::MaxSeqNumber)
    {
        return false;
    }
    _seqNum++;
    return true;
}

bool 
TRIPRouteElement::sameAttribute(const TRIPRouteElement& other) const
{
    if(_attrSet.size() != other._attrSet.size()) return false;

    bool retVal = true;
    Sptr<TRIPRouteAttr> attr = 0;
    Sptr<TRIPRouteAttr> attr2 = 0;
    for(AttributeSet::iterator itr = _attrSet.begin(), itr2 = other._attrSet.begin(); 
        itr != _attrSet.end(); itr++, itr2++)
    {
        attr = (*itr);
        attr2 = (*itr2);
        if(!(attr->isEqualAttribute(*attr2))) 
        {
            return (false);
        }
    }
    return retVal;
}

void 
TRIPRouteElement::removeAttribute(TRIPAttrType aType)
{
    Sptr<TRIPRouteAttr> retAttr = 0;
    for(AttributeSet::iterator itr = _attrSet.begin(); 
                      itr != _attrSet.end(); itr++)
    {
        if((*itr)->getType() == aType)
        {
            _attrSet.erase(itr);
            break;
        }
    }
}

bool
TRIPRouteElement::qualifies(int i) const
{
    return (_aggrStatus & i);
}


bool operator<(const Sptr<TRIPRouteElement>& left,                                                            const Sptr<TRIPRouteElement>& right)
{
     return (*left < *right);
}


