#ifndef TRIPRouteElement_hxx
#define TRIPRouteElement_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const TRIPRouteElement_hxx_Version = 
"$Id: TRIPRouteElement.hxx,v 1.27 2001/03/20 23:01:20 kevinmc Exp $";

#include <netinet/in.h>
#include <sys/types.h>

#include "vtypes.h"
#include "TRIPItadKey.hxx"
#include "TRIPRoutes.hxx"
#include "Data.hxx"
#include "TRIPUpdateMsg.hxx"
#include "TRIPDef.hxx"
#include "LockHelper.hxx"
#include "VMutex.h"

typedef enum
{
    T_AGGR_BIT=0x01,
    T_UPDATE_BIT=0x02,
    T_PROXY_BIT=0x04,
    T_OUTGOING_BIT=0x08,
    T_FLOOD_BIT=0x10,
    T_EXTTRIB_BIT=0x20,
    T_ALL_BIT=T_AGGR_BIT|T_UPDATE_BIT|T_PROXY_BIT|T_OUTGOING_BIT|T_FLOOD_BIT|T_EXTTRIB_BIT
} RouteChangeBitMask;

/** 
  A TRIPRouteElement represents a route in the routing table.Based on the 
  routing attributes the object can be extended.Currently 
  it uses attributes for Load and the cost metrics.
 */
class TRIPRouteElement {
    public:
        ///
        TRIPRouteElement(const TRIPItadKey& key,
                  const TRIPRoutes& route,
                  const Data& nextHopServer,
                  const TRIPUpdateMsg::AttrSet attrSet,
                  bool itrip = false);

        ///
        TRIPRouteElement(const TRIPItadKey& key,
                  const TRIPRoutes& route,
                  const Data& nextHopServer,
                  bool itrip = false);

        ///
        TRIPRouteElement(const TRIPRouteElement& other) ;

        ///
        TRIPRouteElement& operator=(const TRIPRouteElement& other);

        ///
        bool operator==(const TRIPRouteElement& other) const;


        ///
        bool operator<(const TRIPRouteElement& other) const;

        /** @name Accessors */
        //@{

        ///
        bool isChanged(RouteChangeBitMask bit) const
        { 
            LockHelper lck(_mutex);
            return (_changed & bit); 
        };

        ///
        bool isWithdrawn() const
        { 
            LockHelper lck(_mutex);
            return (_withdrawn == 1); 
        };

        ///
        bool isInternal() const { return (_itrip == 1); };

        ///
        u_int32_t getSeqNumber() const { return _seqNum; };


        ///
        const string& getPrefix() const { return _route.getPrefix(); };

        ///
        TRIPAddrFamily getAddrFamily() const { return _route.getAddrFamily(); };

        ///
        TRIPAppProto getAppProtocol() const { return _route.getAppProtocol(); }

        ///
        const Data& getNextHop() const { return _nextHop; };

        ///
        const TRIPItadKey& getSource() const { return _source; };

        ///
        const TRIPItadKey& getOriginator() const { return _originator; };

        ///
        const TRIPRoutes& getRoute() const { return _route; };

        ///
        const Sptr<TRIPRouteAttr> getAttribute(TRIPAttrType aType) const;

        ///
        bool isNonAggregated() const { return(_aggrStatus == T_NON_AGGREGATED); };
        ///
        bool isAggregated() const { return(_aggrStatus == T_AGGREGATED); };

        ///
        bool isAggregatedResult() const { return(_aggrStatus == T_AGGREGATED_RES); };
        ///
        u_int32_t  getTransactionId() const  { return(_trId); };

        ///
        bool   qualifies(int i) const;

        //@}


        /** @name Mutators */
        //@{

        ///
        void addAttribute(Sptr<TRIPRouteAttr> attr);

        ///
        void removeAttribute(TRIPAttrType aType);

        ///
        void setSeqNumber(u_int32_t seqNum) { _seqNum = seqNum; };

        ///
        void setWithdrawn(bool flg=true) 
        { 
            LockHelper lck(_mutex);
            _withdrawn = flg; 
        };

        ///
        void setChanged(RouteChangeBitMask bit) 
        { 
            LockHelper lck(_mutex);
            _changed = static_cast<RouteChangeBitMask>(_changed | bit); 
        };

        ///
        void unsetChanged(RouteChangeBitMask bit) 
        { 
            LockHelper lck(_mutex);
            if (bit == T_ALL_BIT) _changed = static_cast<RouteChangeBitMask>(bit);
            if (_changed & bit) _changed = static_cast<RouteChangeBitMask>(_changed ^ bit);
        };

        /**Returns true if could increment the seq number without reaching
          * MaxSeqNumber.Returns false if the number reaches MaxSeqNumber
        */
        bool incrementSeqNum();

        ///
        bool sameAttribute(const TRIPRouteElement& other ) const;

        ///
        void setOriginator(const TRIPItadKey& orig) { _originator = orig; };

        ///
        void setAggrStatus(TAggrStatus st) 
        { 
            LockHelper lck(_mutex);
            _aggrStatus = st; 
        };

        ///
        void setTransactionId(u_int32_t trId) 
        { 
            LockHelper lck(_mutex);
            _trId = trId; 
        };

        //@}

        ///
        void writeData(int id) const throw(VNetworkException&);

        ///
        friend ostream& operator<<(ostream& os, const TRIPRouteElement&);

        ///
        friend bool operator<(const Sptr<TRIPRouteElement>& left, 
                              const Sptr<TRIPRouteElement>& right);


        ///
        const AttributeSet& getAttributes() const { return _attrSet; };

        ///
        void prettyDump(ostream& os) const;

    private:

        ///
        void copyObj(const TRIPRouteElement& other);


        ///
        TRIPRouteElement();

        ///
        time_t     _timeStamp;
        ///
        u_int32_t  _trId;

        ///
        TRIPRoutes    _route;
        ///
        Data          _nextHop;
        ///
        float         _cost;
        ///
        u_int32_t     _load;
        ///
        TRIPItadKey   _source;
        ///
        TRIPItadKey   _originator;
        ///
        u_int32_t     _seqNum;

        ///
        AttributeSet  _attrSet;

        ///
        RouteChangeBitMask   _changed;
        ///
        u_int8_t      _withdrawn;
        ///
        u_int8_t      _itrip;

        ///
        mutable VMutex          _mutex;

        ///
        TAggrStatus   _aggrStatus;

};

#endif
