
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const TRIPTrieCache_cxx_Version = 
"$Id: TRIPTrieCache.cxx,v 1.18 2001/03/20 23:02:53 kevinmc Exp $";
#include <stdio.h>

#include "TRIPTrieCache.hxx"
#include "cpLog.h"
#include <algorithm>


TRIPTrieCache::TRIPTrieCache(int order)
   : _cType(TC_LOCAL),
     _updated(false),
     _order(order)
{
    init();
} ;

TRIPTrieCache::TRIPTrieCache(const TRIPItadKey& key, TCacheType cType, 
                             int order)
      : _itadKey(key),
        _cType(cType),
        _updated(false),
        _order(order)
{
    init();
};

void
TRIPTrieCache::init()
{
    //create empty nodes and empty route set for each protocol
    //supported
    for(int i = TRIP_APP_PROTO_SIP; i < TRIP_APP_PROTO_MAX;i++)
    {
        TRIPAppProto proto = static_cast<TRIPAppProto>(i);
        TRIPProtoKey key(proto, TRIP_ADDR_POTS);
        _mapOfRouteSet[key] = new TRIPRouteSet;
        _mapOfRoutes[key].setOrder(_order);

        TRIPProtoKey key2(proto, TRIP_ADDR_ROUTING_NUM);
        _order = 16;
        _mapOfRouteSet[key2] = new TRIPRouteSet;
        _mapOfRoutes[key2].setOrder(_order);
    } 
}

void
TRIPTrieCache::clearCache()
{
    for(MapOfRoutes::iterator itr = _mapOfRoutes.begin();
               itr != _mapOfRoutes.end(); itr++)
    {
        (*itr).second.clear();
    }
}

void 
TRIPTrieCache::insert(Sptr<TRIPRouteElement> rec) 
{
    TRIPProtoKey key(rec->getAppProtocol(), rec->getAddrFamily());

    // if route is not inserted then a match exists, remove it and insert again
    pair<TRIPRouteSet::iterator, bool> result = _mapOfRouteSet[key]->insert(rec);
    if (result.second == false)
    {
        _mapOfRouteSet[key]->erase(result.first);
        _mapOfRouteSet[key]->insert(rec);
    }
    _mapOfRoutes[key].add(rec);
    _updated = true;
}

void
TRIPTrieCache::withdraw(Sptr<TRIPRouteElement> rec) 
{
    rec->setWithdrawn(true);
    _updated = true;
}

void 
TRIPTrieCache::remove(Sptr<TRIPRouteElement> rec)
{
    TRIPProtoKey key(rec->getAppProtocol(), rec->getAddrFamily());

    if(!_mapOfRoutes.count(key))
    {
        return;
    }
    Sptr<TRIPRouteSet> rSet = _mapOfRouteSet[key];
    TRIPRouteSet::iterator itr = rSet->find(rec);
    if(itr != rSet->end()) rSet->erase(itr);

    _mapOfRoutes[key].remove(rec->getPrefix());
}


void 
TRIPTrieCache::readLock() const
{
    cpLog(LOG_DEBUG, "Locking trie cache for read.");
    _lock.ReadLock();
}

void 
TRIPTrieCache::writeLock() const
{
    cpLog(LOG_DEBUG, "Locking trie cache for write.");
    _lock.WriteLock();
}

void 
TRIPTrieCache::unlock() const
{
    cpLog(LOG_DEBUG, "Unlocking trie cache.");
    _lock.Unlock();
}


string 
TRIPTrieCache::getNextHopServer(const string& prefix, TRIPProtoKey protoKey) const
{
    Sptr<TRIPRouteElement> rElem = _mapOfRoutes[protoKey].getData(prefix);
    string retStr;
    if(rElem.getPtr() != 0) 
    {
        retStr = rElem->getNextHop().getData();
    }
    else
    {
        cpLog(LOG_INFO, "Did not find route for prefix (%s)", prefix.c_str());
    }
    return retStr;
}

Sptr<TRIPRouteSet>
TRIPTrieCache::getSortedRouteSet(TRIPProtoKey proto) const
{
    Sptr<TRIPRouteSet> rSet = _mapOfRouteSet[proto];
    Sptr<TRIPRouteSet> retSet = new TRIPRouteSet();
//    vector<Sptr<TRIPRouteElement> >  vSet(rSet->begin(), rSet->end());

    vector<Sptr<TRIPRouteElement> >  vSet;
    for (TRIPRouteSet::const_iterator itr = rSet->begin();
                                        itr != rSet->end(); itr++)
    {
        vSet.push_back((*itr));
    }

    sort(vSet.begin(), vSet.end(), GreaterTrid());
    retSet->insert(vSet.begin(), vSet.end()); 
    return retSet;
}

Sptr<TRIPRouteSet> 
TRIPTrieCache::getRouteSet(int flags, TRIPProtoKey protoKey) const
{
    Sptr<TRIPRouteSet> retSet = 0;
    
    if ((flags & T_NON_AGGREGATED) && (flags & T_AGGREGATED) && 
                                            (flags & T_AGGREGATED_RES))
    {
        //everything
        return (_mapOfRouteSet[protoKey]);
    }
    else
    {
        retSet = new TRIPRouteSet();
        Sptr<TRIPRouteSet> rSet = _mapOfRouteSet[protoKey];
        for(TRIPRouteSet::iterator itr = rSet->begin();
                                   itr != rSet->end(); itr++)
        {
            Sptr<TRIPRouteElement> element = (*itr);
            if(element->qualifies(flags)) retSet->insert(element);
        }
    }
    return retSet;
}

Sptr<TRIPRouteSet>
TRIPTrieCache::getNextOutRouteSet(bool& more, TRIPProtoKey protoKey) const
{
    Sptr<TRIPRouteSet> retSet = 0;
    retSet = new TRIPRouteSet();
    int count = 0;
    more = false;
    Sptr<TRIPRouteSet> rSet = _mapOfRouteSet[protoKey];
    for(TRIPRouteSet::iterator itr = rSet->begin();
                                            itr != rSet->end(); itr++)
    {
        if (((*itr)->isChanged(T_OUTGOING_BIT)) && (!(*itr)->isAggregated()))
        {
            (*itr)->unsetChanged(T_OUTGOING_BIT);
            retSet->insert((*itr));
            if (++count >= 10) 
            {
                more = true;
                break;
            }
        }
    }
    return retSet;
}


bool 
TRIPTrieCache::matchDestination(Sptr<TRIPRouteElement> route,                                                         TRIPRouteSet& retSet) const
{
    const string& prefix = route->getPrefix();
    TRIPProtoKey key(route->getAppProtocol(), route->getAddrFamily());
    Sptr<TRIPRouteElement> rElem = 
                 _mapOfRoutes[key].getDataExact(prefix);
    if(rElem.getPtr() != 0)
    {
        retSet.insert(rElem);
    }
    return ((retSet.size()==0) ? false : true);
}

bool 
TRIPTrieCache::moreSpecific(Sptr<TRIPRouteElement> left, TRIPRouteSet& retSet, bool& sameAttribute) const
{
    bool moreSpecificFlg = false;
    const string& prefix = left->getPrefix();
    TRIPProtoKey key(left->getAppProtocol(), left->getAddrFamily());

    Sptr<TRIPRouteElement> rElem = 
             _mapOfRoutes[key].getMoreSpecificTo(prefix, moreSpecificFlg);
    if(moreSpecificFlg)
    {
        if(rElem.getPtr() != 0) 
        {
            sameAttribute = rElem->sameAttribute(*left);
            retSet.insert(rElem);
        }
        else
        {
            //Since this is the only route which is present in the database
            //Nothing to compare
            sameAttribute = true;
        }
    }
    return moreSpecificFlg; 
}

bool
TRIPTrieCache::isInternalRouteCache() const
{
    return (_cType == TC_INTERNAL);
}

bool
TRIPTrieCache::isLocalRouteCache() const
{
    return (_cType == TC_LOCAL);
}

bool
TRIPTrieCache::isExternalRouteCache() const
{
    return (_cType == TC_EXTERNAL);
}

Sptr<TRIPRouteSet> 
TRIPTrieCache::getWithdrawnRoutesChangedNeedsFlooding(TRIPProtoKey protoKey) const
{
    ///Break the set in grouped by the group ID that would make
    ///sure that routes which were part of original Update message
    ///stay together to keep the attribute consistent.
    Sptr<TRIPRouteSet> retSet = new TRIPRouteSet();
    TRIPRouteSet::iterator itr ;
    ///Get the first element that changed and get its group id that would
    ///be used to group all elements that changed and belong to that group
    ///Function can be called in a loop till all the groups are return.Return null
    ///When no groups are left
    bool firstElem = true;
    u_int32_t trId = 0;
    Sptr<TRIPRouteSet> rSet = _mapOfRouteSet[protoKey];
    for (itr = rSet->begin(); itr != rSet->end(); itr++)
    {
        Sptr<TRIPRouteElement> data = (*itr);
        if(data->isChanged((RouteChangeBitMask)(T_UPDATE_BIT|T_FLOOD_BIT)) && !data->isWithdrawn())
        {
            if(firstElem) 
            {
                trId = data->getTransactionId();
                firstElem = false;
            }
            if(trId == data->getTransactionId()) retSet->insert(data);
        }
    }
    return retSet;
}

Sptr<TRIPRouteSet> 
TRIPTrieCache::getReachableRoutesChangedNeedsFlooding(u_int32_t trId, TRIPProtoKey protoKey) const
{
    Sptr<TRIPRouteSet> retSet = new TRIPRouteSet();
    TRIPRouteSet::iterator itr ;
    Sptr<TRIPRouteSet> rSet = _mapOfRouteSet[protoKey];
    for (itr = rSet->begin(); itr != rSet->end(); itr++)
    {
        Sptr<TRIPRouteElement> data = (*itr);
        if(data->isChanged((RouteChangeBitMask)(T_UPDATE_BIT|T_FLOOD_BIT)) && 
           (!data->isWithdrawn()) && (data->getTransactionId() == trId))
        {
             retSet->insert(data);
        }
    }
    return retSet;
}

Sptr<TRIPRouteSet> 
TRIPTrieCache::getReachableRoutesChangedNeedsFlooding(TRIPProtoKey protoKey) const
{
    Sptr<TRIPRouteSet> retSet = new TRIPRouteSet();
    TRIPRouteSet::iterator itr ;
    ///Get the first element that changed and get its group id that would
    ///be used to group all elements that changed and belong to that group
    ///Function can be called in a loop till all the groups are return.Return null
    ///When no groups are left
    bool firstElem = true;
    u_int32_t trId = 0;
    Sptr<TRIPRouteSet> rSet = _mapOfRouteSet[protoKey];
    for (itr = rSet->begin(); itr != rSet->end(); itr++)
    {
        Sptr<TRIPRouteElement> data = (*itr);
        if(data->isChanged((RouteChangeBitMask)(T_UPDATE_BIT|T_FLOOD_BIT)) && (!data->isWithdrawn()))
        {
            if(firstElem) 
            {
                trId = data->getTransactionId();
                firstElem = false;
            }
            if(trId == data->getTransactionId()) retSet->insert(data);
        }
    }
    return retSet;
}

bool 
TRIPTrieCache::isAggregatable(const string& prefix, string& parentPrefix,
                                Sptr<TRIPRouteElement>& route, 
                                TRIPRouteSet& retSet, TRIPProtoKey protoKey)
{
    if (_mapOfRoutes[protoKey].isAggregatable(prefix, parentPrefix, route, retSet))    
    {
        for(TRIPRouteSet::const_iterator itr = retSet.begin();
                                                itr != retSet.end(); itr++)
        {
            if ((*itr)->isWithdrawn())
            {
                return false;
            }
        }
        return true;
    }
    return false;
}

void 
TRIPTrieCache::removeElementWhereSourceIs(const TRIPItadKey& key)
{
    for(MapOfRouteSet::iterator itr = _mapOfRouteSet.begin();
            itr != _mapOfRouteSet.end(); itr++)
    {
        Sptr<TRIPRouteSet> rSet = (*itr).second;
        for(TRIPRouteSet::iterator itr = rSet->begin();
                                   itr != rSet->end(); itr++)
        {
            if((*itr)->getSource() == key)
            {
                remove(*itr);
            }
        }
    }
}

void
TRIPTrieCache::prettyDump(ostream& os) const
{
    for( TRIPTrieCache::MapOfRouteSet::const_iterator itr = _mapOfRouteSet.begin();
                                                    itr != _mapOfRouteSet.end(); itr++)
    {
        os.width(25);
        os << "Protocol- " << (*itr).first << endl;
        Sptr<TRIPRouteSet> routeSet = (*itr).second;
        for (TRIPRouteSet::const_iterator rteItr = routeSet->begin();
                                            rteItr != routeSet->end(); rteItr++)
        {
            (*rteItr)->prettyDump(os);
        }
    }
}
    
