#ifndef TRIPTrieCache_hxx
#define TRIPTrieCache_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const TRIPTrieCache_hxx_Version = 
"$Id: TRIPTrieCache.hxx,v 1.12 2001/03/06 03:02:31 kevinmc Exp $";

#include <set>
#include "vtypes.h"
#include "Sptr.hxx"
#include "VRWLock.hxx"
#include "TRIPRouteElement.hxx"
#include "TRIPTrieNode.hxx"
#include "TRIPProtoKey.hxx"
#include "TRIPDef.hxx"

/** TRIPTrieCache implements the Routing database which stores the routes
    in a Trie tree structure.
    <pre>
    The cache currently support 
    1. Trie of order 10 - The default and expects characters in the 
       range [0-9] for POTS numbers.
    2. Trie of order 16 - And expects characters in the ranges
       [0-9] and [A-F] for routing numbers.

    Uses private object TRIPTrieNode to implement the Trie.
    Cache partitions the routing information based on the AddressFamily 
    (POTS and RoutingNumber) and protocol type.A TRIPProtoKey, which
    is a composite key of AddressFamily and Protocol, is used to access
    routing information in the cache.
    </pre>
*/
class TRIPTrieCache 
{
   public:
      ///
      TRIPTrieCache(int order=10);
     
      ///
      TRIPTrieCache(const TRIPItadKey& key, TCacheType cType, int order=10);

      ///
      void clearCache();

      ///
      void insert(Sptr<TRIPRouteElement> rElem) ;

      ///
      void withdraw(Sptr<TRIPRouteElement> rElem) ;

      ///
      bool isUpdated() const { return _updated; };

      ///
      void setUpdated(bool flg=true) { _updated = flg; };

      ///Returns the closest match to the pattern
      string getNextHopServer(const string& prefix, TRIPProtoKey proto) const;

      ///Flags is an OR of T_NON_AGGREGATED, T_AGGREGATED and T_AGGREGATED_RES
      Sptr<TRIPRouteSet> getRouteSet(int flags, TRIPProtoKey proto)  const;

      ///Get the routeset in sorted order of Transaction Id for initial flooding
      Sptr<TRIPRouteSet> getSortedRouteSet(TRIPProtoKey proto) const;

      ///Get entire route set
      Sptr<TRIPRouteSet> getRouteSet(TRIPProtoKey proto)  const
      {
          return(getRouteSet(T_NON_AGGREGATED| T_AGGREGATED | T_AGGREGATED_RES, proto));
      }

      ///Get a set of routes with T_OUTGOING_BIT set
      Sptr<TRIPRouteSet> getNextOutRouteSet(bool& more, TRIPProtoKey proto) const;
     
      ///
      bool matchDestination(Sptr<TRIPRouteElement> route, 
                                  TRIPRouteSet& retSet) const; 

      ///
      void readLock() const;

      /**Finds the route that is more specific that left and returns it in
      *  retSet.Also if the left has the same attribute return true in the 
      *  sameAttribute
      */
      bool moreSpecific(Sptr<TRIPRouteElement> left, TRIPRouteSet& retSet, bool& sameAttribute) const;

      ///
      void writeLock() const;
      ///
      void unlock() const;

      ///
      friend ostream& operator<<(ostream& os, const TRIPTrieCache& obj);

      ///Returns true if it is a Internal cache
      bool isInternalRouteCache() const;
      
      ///Returns true if it is a Local Route cache
      bool isLocalRouteCache() const;

      ///Returns true if it is a External Route cache
      bool isExternalRouteCache() const;

      ///
      const TRIPItadKey&  getItadKey() const { return _itadKey; };

      /**Returns the first group of withdrawn elements that changed and needs to
       * be flooded
       * Get the first element that changed and get its group id that would
       * be used to group all elements that changed and belong to that group
       * Function can be called in a loop till all the groups are return.Return null
       * When no groups are left.
       */
      Sptr<TRIPRouteSet> getWithdrawnRoutesChangedNeedsFlooding(TRIPProtoKey proto) const;

      /**Return the set of reachable routes that belong to a specific group and
       * needs to be flooded.
       */
      Sptr<TRIPRouteSet> getReachableRoutesChangedNeedsFlooding(u_int32_t trId, TRIPProtoKey proto) const;

      /**Returns the first group of reachable elements that changed and needs to be
       * flooded.
       * Get the first element that changed and get its group id that would
       * be used to group all elements that changed and belong to that group
       * Function can be called in a loop till all the groups are return.Return null
       * When no groups are left.
       */
      Sptr<TRIPRouteSet> getReachableRoutesChangedNeedsFlooding(TRIPProtoKey proto) const;

      /** Checks to see if a given prefix is aggregatable, if so returns true
        * and also returns the ref. to the parent prefix, route and all children
        * to the parent
        */
      bool isAggregatable(const string& prefix, string& parentPrefix,
                                Sptr<TRIPRouteElement>& route, 
                                TRIPRouteSet& retSet, 
                                TRIPProtoKey proto);

      ///
      void removeElementWhereSourceIs(const TRIPItadKey& key);

      ///
      void prettyDump(ostream& os) const;

      ///
      void remove(Sptr<TRIPRouteElement> rec);
    private:

      ///
      void init();

      ///For each protocol there is a separate tri tree
      typedef map<TRIPProtoKey, TRIPTrieNode<TRIPRouteElement> > MapOfRoutes;
      ///
      mutable MapOfRoutes _mapOfRoutes;
      //TRIPTrieNode<TRIPRouteElement> _root;

      ///
      TRIPItadKey  _itadKey;

      ///
      TCacheType   _cType;

      mutable bool   _updated;
      ///
      mutable VRWLock  _lock;

      ///
      int _order;

      ///
      typedef map<TRIPProtoKey, Sptr<TRIPRouteSet> > MapOfRouteSet;
      ///
      mutable MapOfRouteSet  _mapOfRouteSet;
};

inline ostream& operator<<(ostream& os, const TRIPTrieCache& obj)
{
    os.width(25);
    os << "**TRIPTrieCache**" << endl;
    for( TRIPTrieCache::MapOfRoutes::const_iterator itr = obj._mapOfRoutes.begin();
                    itr != obj._mapOfRoutes.end(); itr++)
    {
        os << (*itr).second << endl;
    }
    return os;
}

/** GreaterTrid is a comparison function class which is used to sort the RouteSet
    of a TrieCache in Transaction Id order
    </pre>
*/
class GreaterTrid : public binary_function<Sptr <TRIPRouteElement> , Sptr<TRIPRouteElement> , bool >
{
    public:
        bool operator() (Sptr <TRIPRouteElement> r1, 
                                        Sptr <TRIPRouteElement> r2)
        {
            return (r1->getTransactionId() < r2->getTransactionId());
        }
};


#endif
