/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const TRIPErrorHandler_cxx_Version = 
"$Id: TRIPErrorHandler.cxx,v 1.6 2001/01/16 03:26:01 sprajpat Exp $";

#include <errno.h>
#include <stdio.h>
#include <strstream.h>
#include <unistd.h>
#include <stdlib.h>

#include "TRIPErrorHandler.hxx"
#include "cpLog.h"

static int MAX_HDR_MSG_CODE=2;
static int MAX_UPDATE_MSG_CODE=6;
static int MAX_OPEN_MSG_CODE=6;

static const char* eMsg[] = 
{
    "Undefined",
    "Message Header Error",
    "Open Message Error",
    "Update Message Error",
    "Hold Timer Expired",
    "Finite State Machine Error",
    "Cease",
    0
};

static const char* eMsgHdrMsg[] = 
{
    "Undefined",
    "Bad Message Length",
    "Bad Message Type",
    0
};

static const char* eOpenMsg[] = 
{
    "Undefined",
    "Unsupported Version Number",    
    "Bad Peer ITAD",    
    "Bad TRIP Identifier",    
    "Unsupported Optional Parameter",    
    "Unacceptable Hold Time",    
    "Unsupported Capability",
    0
};

static const char* eUpdateMsg[] = 
{
    "Undefined",
    "Malformed Attribute List",    
    "Unrecognized Well-known Attribute",    
    "Missing Well-known Mandatory Attribute",    
    "Attribute Flags Error",    
    "Attribute Length Error",    
    "Invalid Attribute",
    0
};


TRIPErrorHandler* TRIPErrorHandler::_instance = 0;

TRIPErrorHandler& 
TRIPErrorHandler::instance()
{
    if(!_instance)
    {
        _instance = new TRIPErrorHandler();
        cpLog(LOG_DEBUG, "Created instance for %s\n", _instance->className().c_str());
    }
    return *_instance;

}

TRIPErrorHandler::TRIPErrorHandler()
{
}

TRIPErrorHandler::~TRIPErrorHandler()
{
    //TODO
}

void 
TRIPErrorHandler::validateCodes(u_int8_t eCode, u_int8_t esubCode) 
    throw (TRIPBadDataException& )
{
    //Validate the error code and errorSubode
    bool failed = false;
    int eSubCode = esubCode;
    switch(static_cast<TRIPErrorCode>(eCode))
    {
        case TRIP_ERR_MSG_HEADER:
        {
            if(eSubCode < 1 || eSubCode > MAX_HDR_MSG_CODE)
            {
                failed = true;
            }
        }
        break;
        case TRIP_ERR_OPEN:
        case TRIP_ERR_UPDATE:
        {
            if(eSubCode < 1 || eSubCode > MAX_UPDATE_MSG_CODE)
            {
                failed = true;
            }
        }
        break;
        case TRIP_ERR_HOLD_TIMER:
        case TRIP_ERR_FSM:
        case TRIP_ERR_CEASE:
        {
            if(eSubCode  != 0)
            {
                failed = true;
            }
        }
        break;
        default:
        {
            failed = true;
        }
        break;
    }
    if(failed)
    {
        char buf[256];
        sprintf(buf, "Bad Error code (%d) or subcode (%d)", 
                              (int)eCode, (int)eSubCode);
        cpLog(LOG_ALERT, buf);
        throw TRIPBadDataException(buf, __FILE__, __LINE__, 0);
    }
}

string
TRIPErrorHandler::getErrorStr(u_int8_t eCode, u_int8_t eSubCode)
     throw (TRIPBadDataException&)
{
    validateCodes(eCode, eSubCode);
    string eStr = eMsg[eCode];
    string eSubStr = getSubcodeStr(eCode, eSubCode);
    return (eStr + "," + eSubStr); 
}



string 
TRIPErrorHandler::getSubcodeStr(u_int8_t eCode, u_int8_t eSubCode) 
     throw (TRIPBadDataException&)
{
    string retStr;
    validateCodes(eCode, eSubCode);
    switch(static_cast<TRIPErrorCode>(eCode))
    {
        case TRIP_ERR_MSG_HEADER:
            retStr = eMsgHdrMsg[eSubCode]; 
        break;
        case TRIP_ERR_OPEN:
            retStr = eOpenMsg[eSubCode]; 
        break;
        case TRIP_ERR_UPDATE:
            retStr = eUpdateMsg[eSubCode]; 
        break;
        case TRIP_ERR_HOLD_TIMER:
        case TRIP_ERR_FSM:
        case TRIP_ERR_CEASE:
        default:
            retStr = "Unknown";
        break;
    }
    return retStr;
}
