/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const TRIPOpenMsg_cxx_Version = 
"$Id: TRIPOpenMsg.cxx,v 1.8 2001/03/05 20:21:19 kevinmc Exp $";

#include <strstream.h>

#include "TRIPOpenMsg.hxx"
#include "TRIPUtil.hxx"
#include "TRIPErrorHandler.hxx"
#include "cpLog.h"

static const int OPEN_MSG_MIN_LENGTH=15;
static struct shipIt {
    u_int8_t    version;
    u_int8_t    res;
    u_int16_t   itad;
    u_int32_t   tripId;
    u_int16_t   hTime;
    u_int16_t   pLen;
} sIt;

TRIPOpenMsg::TRIPOpenMsg(u_int16_t itadNumber, 
                         u_int32_t tripId,u_int16_t holdTime)
    :TRIPMsg(), 
     _version(1), 
     _res(0),
     _itad(itadNumber), 
     _tripId(tripId), 
     _holdTime(holdTime),  
     _paramLen(0)
{
    _hdr = new TRIPHeader(TRIP_OP_OPEN);
    //Length of message without any parameters is 15 bytes
    _hdr->setMsgLength(OPEN_MSG_MIN_LENGTH);
}

void 
TRIPOpenMsg::dump(ostream& os)const
{
   os.width(DUMP_LABEL_WIDTH);
   os << "**OpenMsg**" << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << *_hdr << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << "Version:" << static_cast<int>(_version) << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << "ITAD:" << static_cast<int>(_itad) << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << "TRIPID:" << _tripId << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << "HoldTime(sec):" << _holdTime << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << "Param length:" << _paramLen << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << "Optional Parameters:" << endl;
   for(ParameterList::const_iterator itr = _parameterList.begin(); 
             itr != _parameterList.end(); itr++)
   {
       os << endl << *(*itr) << endl;
   }
}

void 
TRIPOpenMsg::parse(const TRIPHeader& hdr,
                          const char* data) throw (TRIPBadDataException&)
{
    _hdr = new TRIPHeader(hdr);

    int idx = 0;
    memcpy(&sIt, data , sizeof(struct shipIt)); 
    idx += sizeof(struct shipIt);
    _version = sIt.version;
    _res = sIt.res;
    _itad = ntohs(sIt.itad);
    _tripId = TRIPId(ntohl(sIt.tripId));
    _holdTime = ntohs(sIt.hTime);
    int pLen = ntohs(sIt.pLen);

    _paramLen = 0;

    u_int16_t tLen = 0;
    //Loop the the entire length of OPEN message and create
    //the sub-component of it.
    while(tLen < pLen)
    {
        u_int16_t type;
        u_int16_t len;
        memcpy(&type, data+idx, 2);
        memcpy(&len, data+idx+2, 2);
        type = ntohs(type);
        len = ntohs(len);
        switch(static_cast<TRIPParamType>(type))
        {
            case TRIP_PARAM_CAP:
            {
                Sptr<TRIPCapabilityInfo> cInfo = new TRIPCapabilityInfo(data+idx); 
                addParameter(cInfo);
                idx += cInfo->getParamLength();
                tLen += cInfo->getLength();
                assert(len == cInfo->getParamLength());
            }
            break;
            default:
                cpLog(LOG_DEBUG, "Unsupported param type (%d)", type);
                tLen += len;
                break;
            break;
        }  
    }
    checkSanity();
}

void 
TRIPOpenMsg::parse(const char* data) throw (TRIPBadDataException&) 
{
    assert(data);
    TRIPHeader hdr(data);
    parse(hdr, data+hdr.getHdrLength());
}


void 
TRIPOpenMsg::writeData(int id) const throw(VNetworkException&)
{
   TRIPMsg::writeData(id);
   sIt.version = _version;
   sIt.res = _res;
   sIt.itad = htons(_itad);
   sIt.tripId = htonl(_tripId.getId());
   sIt.hTime = htons(_holdTime);
   sIt.pLen = htons(_paramLen);

   TRIPUtil::writeData(id, &sIt, sizeof(struct shipIt));

   for(ParameterList::const_iterator itr = _parameterList.begin(); 
             itr != _parameterList.end(); itr++)
   {
       (*itr)->writeData(id);
   }
}

void
TRIPOpenMsg::addParameter(Sptr<TRIPParameter> parameter)
{
    assert(parameter.getPtr() != 0);
    _parameterList.push_back(parameter);
    _paramLen += parameter->getLength();
    setMsgLength(_paramLen + OPEN_MSG_MIN_LENGTH);
}


void 
TRIPOpenMsg::checkSanity() const throw (TRIPBadDataException&)
{
    if((!_hdr->isAOpenHdr()) && (_hdr->getMsgLength() < OPEN_MSG_MIN_LENGTH))
    {
        throw TRIPBadDataException("Bad Open Msg", __FILE__, __LINE__, 
                                                            TRIP_ERR_OPEN);
    }
}

Sptr<TRIPTransCapability>
TRIPOpenMsg::getTransCapability() const throw (TRIPBadDataException&)
{
    Sptr<TRIPTransCapability> retVal=0;
    ParameterList::const_iterator itr; 
    for (itr = _parameterList.begin(); itr != _parameterList.end(); itr++)
    {
        const Sptr<TRIPParameter>& param = (*itr);
        if(param->getType() == TRIP_PARAM_CAP)
        {
            Sptr<TRIPCapabilityInfo> cInfo;
            cInfo.dynamicCast(param);
            assert(cInfo.getPtr());
            const TRIPCapabilityInfo::CapabilityList& cList = cInfo->getData();
            TRIPCapabilityInfo::CapabilityList::const_iterator cItr;
            for (cItr = cList.begin(); cItr != cList.end(); cItr++)
            {
                const Sptr<TRIPCapability>& cap = (*cItr);
                if(cap->getCode() == TRIP_CAP_SND_REC)
                {
                    Sptr<TRIPTransCapability> transCap ;
                    transCap.dynamicCast(cap);
                    assert(transCap.getPtr());
                    return(transCap);
                }
             }
         }
     }
     ///Could not find tranCapability info, error
     throw TRIPBadDataException("Missing Trans Capability",
                      __FILE__, __LINE__, TRIP_ERR_OPEN, O_UNSUPP_VERSION);
}
