#ifndef TRIPRoutes_hxx
#define TRIPRoutes_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const TRIPRoutes_hxx_Version = 
"$Id: TRIPRoutes.hxx,v 1.10 2001/01/23 02:11:44 sprajpat Exp $";

#include "vtypes.h"
#include "VNetworkException.hxx"
#include "TRIPBadDataException.hxx"
#include "TRIPRouteHdr.hxx"
#include "TRIPData.hxx"
#include "TRIPCoreDef.hxx"

/** Implements a TRIP route element.The Genegeric format of TRIP route
    <pre>
    +-----------------------------------------------------------------+
    +   TRIPRouteHdr (4 bytes)                                        +
    +-----------------------------------------------------------------+
    +  Length (2 bytes)          |  Address (variable)                +
    +-----------------------------------------------------------------+
    </pre>
*/
class TRIPRoutes {
    public:
        /**Constructor to create a route from a given addr header and given data
         *Usage would be TRIPRoutes(TRIP_ADDR_POTS, TRIP_APP_PROTO_SIP, 
         *                          TRIPData("1408"));
        */
        TRIPRoutes(TRIPAddrFamily aFamily ,
                   TRIPAppProto protocol , const TRIPData& data);

        /**Constructor to create a route from a given addr header and given data
         * Usage : 
         *  TRIPRouteHdr hdr(TRIP_ADDR_POTS, TRIP_APP_PROTO_SIP);
         *  TRIPRoutes(hdr, TRIPData("1408");
         */
        TRIPRoutes(const TRIPRouteHdr& hdr, const TRIPData& data);

        ///
        virtual ~TRIPRoutes() { };

        ///
        TRIPRoutes(const TRIPRoutes& other); 

        ///
        TRIPRoutes& operator=(const TRIPRoutes& other);

        ///
        bool operator==(const TRIPRoutes& other);


        /**@name Accessors */
        //@{
        ///
        u_int16_t getLength() const;

        ///Gets the address family of the route
        inline TRIPAddrFamily getAddrFamily() const { return _routeHdr->getAddrFamily(); }

        ///Gets the application protocol for the route
        inline TRIPAppProto getAppProtocol() const { return _routeHdr->getAppProtocol(); }

        ///Returns the string representation of route address
        inline const string& getPrefix() const { return _data->str(); };

        //@}

        ///
        friend ostream& operator<<(ostream& os, const TRIPRoutes&);

        ///
        inline bool operator<(const TRIPRoutes& other) const
        {
            if(_routeHdr->getAppProtocol() < other._routeHdr->getAppProtocol())
            { 
                return true;
            }
            if(_routeHdr->getAddrFamily() < other._routeHdr->getAddrFamily())
            {
                return true;
            }
            if(*_data < *(other._data))
            {
                return true;
            }
            return false;
        }

        ///Writes object to given network socket in network byte order
        void writeData(int id) const throw(VNetworkException&);

    protected:
        ///
        void processAddress(const TRIPRouteHdr& hdr, const char* buf, int len);

        ///
        TRIPRoutes(): _routeHdr(0), _data(new TRIPData("")) { };

        ///
        TRIPRoutes(const char* buf) throw (TRIPBadDataException&);

        ///
        void copyObj(const TRIPRoutes& other);

        ///
        Sptr<TRIPRouteHdr> _routeHdr;
        ///
        u_int16_t          _length;
        ///
        Sptr<TRIPData>     _data;

        ///
        friend class TRIPRouteContainer;
        ///
        friend class TRIPRouteElement;
};

inline ostream& operator<<(ostream& os, const TRIPRoutes& obj)
{
    os.width(DUMP_LABEL_WIDTH);
    os << "**Routes**" << endl;
    os.width(DUMP_LABEL_WIDTH);
    os << "Route Hdr: " << *(obj._routeHdr) << endl;
    os.width(DUMP_LABEL_WIDTH);
    os << "length: " << static_cast<int>(obj._length) << endl;
    os.width(DUMP_LABEL_WIDTH);
    os << "Data: " << *(obj._data) << endl;
    return os;
}

#endif
