/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const TRIPUpdateMsg_cxx_Version = 
"$Id: TRIPUpdateMsg.cxx,v 1.25 2001/03/09 22:07:06 kevinmc Exp $";

#include <stdio.h>
#include <strstream.h>

#include "TRIPUpdateMsg.hxx"
#include "TRIPUtil.hxx"
#include "TRIPErrorHandler.hxx"
#include "TRIPCoreDef.hxx"
#include "cpLog.h"

TRIPUpdateMsg::TRIPUpdateMsg()
    :TRIPMsg() 
{
    FD_ZERO(&_aFlagSet);
    _hdr = new TRIPHeader(TRIP_OP_UPDATE);
}


TRIPUpdateMsg::TRIPUpdateMsg(Sptr<TRIPWithdrawnRoute> wRoute,
                    Sptr<TRIPReachableRoute> rRoute, const AttrSet& aSet,
                    bool isExternal)
                    :TRIPMsg()
{
    FD_ZERO(&_aFlagSet);
    _hdr = new TRIPHeader(TRIP_OP_UPDATE);
    if (!(wRoute.getPtr()))
    {
        assert(0);
    }
    if (!(rRoute.getPtr()))
    {
        assert(0);
    }
    addWithdrawnRoute(wRoute);
    addReachableRoute(rRoute);
    for(AttrSet::const_iterator itr = aSet.begin();
                 itr != aSet.end() ; itr++)
    {
        if (isExternal)
        {
            if ((*itr)->getType() == TRIP_LOCAL_PREF) continue;
        }
        addAttribute((*itr));
    }
}
    
    

void 
TRIPUpdateMsg::dump(ostream& os)const
{
   os.width(DUMP_LABEL_WIDTH);
   os << "**UpdateMsg**" << endl;
   checkSanity();
   os.width(DUMP_LABEL_WIDTH);
   os << *_hdr << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << *_withdrawnRoute << endl;
   os.width(DUMP_LABEL_WIDTH);
   os << *_reachableRoute << endl;

   for(AttrSet::iterator itr = _attrSet.begin();
                 itr != _attrSet.end() ; itr++)
   {
       os.width(DUMP_LABEL_WIDTH);
       os << *(*itr) << endl;
   }
}

void 
TRIPUpdateMsg::parse(const TRIPHeader& hdr,
                          const char* data) throw (TRIPBadDataException&)
{
    if(!(hdr.isAUpdateHdr()))
    {
        throw TRIPBadDataException("Bad Update Msg", __FILE__, 
                                    __LINE__, TRIP_ERR_MSG_HEADER);
    }

    FD_ZERO(&_aFlagSet);

    _hdr = new TRIPHeader(hdr);

    int tLen = hdr.getMsgLength() - hdr.getHdrLength();
    assert(tLen > HDR_LENGTH);
    int cLen = 0;
    //Loop through the entire message length and parse the individual 
    //elements of the message.
    while(cLen < tLen)
    {
        TRIPAttrHeader::TripAttrHeaderData hData;
        memcpy(&hData, data+cLen, sizeof(TRIPAttrHeader::TripAttrHeaderData));
        u_int8_t type = hData._type;
        u_int16_t len = ntohs(hData._length);
        validateDuplicateAttr(type);
        switch(static_cast<TRIPAttrType>(type))
        {
            case TRIP_WITHDRAWN_ROUTE:
            {
                _withdrawnRoute = new TRIPWithdrawnRoute(data+cLen);
                cLen += _withdrawnRoute->getLength();
                FD_SET(TRIP_WITHDRAWN_ROUTE, &_aFlagSet);
                setMsgLength(cLen + HDR_LENGTH);
                cpLog(LOG_DEBUG, "Created TRIP_WITHDRAWN_ROUTE Attribute ");
            }
            break;
            case TRIP_REACHABLE_ROUTE:
            {
                _reachableRoute = new TRIPReachableRoute(data+cLen);
                cLen += _reachableRoute->getLength();
                FD_SET(TRIP_REACHABLE_ROUTE, &_aFlagSet);
                setMsgLength(cLen + HDR_LENGTH);
                cpLog(LOG_DEBUG, "Created TRIP_REACHABLE_ROUTE Attribute ");
            }
            break;
            case TRIP_NEXT_HOP_SERVER:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPNextHopServer(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
                cpLog(LOG_DEBUG, "Created TRIP_NEXT_HOP_SERVER Attribute ");
            }
            break;
            case TRIP_ADVERTISEMENT_PATH:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPAdvertisePath(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
                cpLog(LOG_DEBUG, "Created TRIP_ADVERTISEMENT_PATH Attribute ");
            }
            break;
            case TRIP_ROUTED_PATH:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPRoutedPath(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
                cpLog(LOG_DEBUG, "Created TRIP_ROUTED_PATH Attribute ");
            }
            break;
            case TRIP_ATOMIC_AGGREGATE:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPAtomicAggregate(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
            }
            break;
            case TRIP_LOCAL_PREF:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPLocalPref(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
            }
            break;
            case TRIP_MULTI_EXIT_DISC:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPMultiExitDisc(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
            }
            break;
            case TRIP_COMMUNITY:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPCommunity(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
            }
            break;
            case TRIP_ITAD_TOPOLOGY:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPItadTopology(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
            }
            break;
            case TRIP_AUTHENTICATION:
            {
                Sptr<TRIPRouteAttr> attr =  new TRIPAuthAttr(data+cLen);
                addAttribute(attr);
                cLen += attr->getLength();
            }
            break;
            default:
                cpLog(LOG_DEBUG, "Unsupported attribute type (%d) with length (%d), ignoring..", type, len);
                cLen += len;
                //If it was a well known attribute, then throw an exception that
                //would send a notification
                if(!(hData._flag & TRIP_FLG_OPT))
                {
                    char msgBuf[256];
                    sprintf(msgBuf, "(%d, %d)", hData._type, hData._length);
                     throw TRIPBadDataException(msgBuf, 
                                 __FILE__, __LINE__, TRIP_ERR_UPDATE, U_UNRECN_ATTR);
                }
            break;
        }
    }
}

void
TRIPUpdateMsg::validateDuplicateAttr(u_int16_t type) 
               throw (TRIPBadDataException&)
{
    if(FD_ISSET(static_cast<int>(type), &_aFlagSet))
    {
        //Attribute of same type occuring twice, bad UpdateMessage
        char mBuf[256];
        sprintf(mBuf, "Malformed Attribute list for type (%d)", static_cast<int>(type));
        throw TRIPBadDataException(mBuf,
                 __FILE__, __LINE__, TRIP_ERR_UPDATE, U_MALFORMED_ATTR_LIST);
    }
}

void 
TRIPUpdateMsg::parse(const char* data) throw (TRIPBadDataException&) 
{
    TRIPHeader hdr(data);
    parse(hdr, data+HDR_LENGTH);
}


void 
TRIPUpdateMsg::writeData(int id) const throw(VNetworkException&)
{
    TRIPMsg::writeData(id);
    _withdrawnRoute->writeData(id);
    _reachableRoute->writeData(id);

    for(AttrSet::iterator itr = _attrSet.begin();
                 itr != _attrSet.end() ; itr++)
    {
        (*itr)->writeData(id);
    }
}

void 
TRIPUpdateMsg::checkSanity() const throw (TRIPBadDataException&)
{
    if(!(_hdr->isAUpdateHdr()))
    {
        throw TRIPBadDataException("Bad update msg header",
                  __FILE__, __LINE__, TRIP_ERR_MSG_HEADER, H_BAD_MSG_TYPE);
    }
    if(_withdrawnRoute.getPtr() == 0)
    {
        cpLog(LOG_ALERT, "Missing mandatory attribute (WithdrawnRoute) in Msg");
        throw TRIPBadDataException("Missing mandatory attribute (withdrawnRoute) in Msg",
                           __FILE__, __LINE__, TRIP_ERR_UPDATE, U_MISSING_MANDATORY_ATTR);
    }
    if(_reachableRoute.getPtr() == 0)
    {
        cpLog(LOG_ALERT, "Missing mandatory attribute (ReachableRoute) in Msg");
        throw TRIPBadDataException("Missing mandatory attribute (ReachableRoute) in Msg",
                                    __FILE__, __LINE__, TRIP_ERR_UPDATE, U_MISSING_MANDATORY_ATTR);
    };
    if(!FD_ISSET(TRIP_NEXT_HOP_SERVER, &_aFlagSet)) 
    {
        cpLog(LOG_ALERT, "Missing mandatory attribute (NextHopServer) in Msg");
        throw TRIPBadDataException("Missing mandatory attribute (NextHopServer) in Msg",
                                    __FILE__, __LINE__, TRIP_ERR_UPDATE, U_MISSING_MANDATORY_ATTR);
    };
    if(!FD_ISSET(TRIP_ADVERTISEMENT_PATH, &_aFlagSet))
    {
        cpLog(LOG_ALERT, "Missing mandatory attribute (AdvertisePath) in Msg");
        throw TRIPBadDataException("Missing mandatory attribute (AdvertisePath) in Msg",
                                    __FILE__, __LINE__, TRIP_ERR_UPDATE, U_MISSING_MANDATORY_ATTR);
    };
    if(!FD_ISSET(TRIP_ROUTED_PATH, &_aFlagSet))
    {
        cpLog(LOG_ALERT, "Missing mandatory attribute (RoutedPath) in Msg");
        throw TRIPBadDataException("Missing mandatory attribute (RoutedPath) in Msg",
                                    __FILE__, __LINE__, TRIP_ERR_UPDATE, U_MISSING_MANDATORY_ATTR);
    };
}


///
void 
TRIPUpdateMsg::addAttribute(Sptr<TRIPRouteAttr> attr)
{
    int len = 0;
    if(FD_ISSET(attr->getType(), &_aFlagSet))
    {
        cpLog(LOG_ALERT, "TRIPUpdateMsg::addAttribute:Attr type (%d) already added, ignoring", attr->getType());
        return;
    }
    FD_SET(attr->getType(), &_aFlagSet);
    _attrSet.insert(attr);
    len = attr->getLength();
    setMsgLength(getMsgLength() + len);
}


void
TRIPUpdateMsg::addReachableRoute(Sptr<TRIPReachableRoute> route)
{
    assert(_reachableRoute.getPtr() == 0);
    _reachableRoute = route;
    int len = getMsgLength();
    len = len + _reachableRoute->getLength();
    setMsgLength(len);
}
       

void
TRIPUpdateMsg::addWithdrawnRoute(Sptr<TRIPWithdrawnRoute> route)
{
    assert(_withdrawnRoute.getPtr() == 0);
    _withdrawnRoute = route;
    int len = getMsgLength();
    len = len + _withdrawnRoute->getLength();
    setMsgLength(len);
}


const set<Sptr<TRIPRouteAttr> > &
TRIPUpdateMsg::getAttributes() const
{
    return _attrSet;
}

bool 
TRIPUpdateMsg::hasAtomicAggregate() const 
{
    return (FD_ISSET(TRIP_ATOMIC_AGGREGATE, &_aFlagSet));
}

bool 
TRIPUpdateMsg::hasLocalPref() const 
{
    return (FD_ISSET(TRIP_LOCAL_PREF, &_aFlagSet));
}

bool 
TRIPUpdateMsg::hasCommunity() const 
{
    return (FD_ISSET(TRIP_COMMUNITY, &_aFlagSet));
}

bool 
TRIPUpdateMsg::hasMultiExitDisc() const 
{
    return (FD_ISSET(TRIP_MULTI_EXIT_DISC, &_aFlagSet));
}

bool 
TRIPUpdateMsg::hasItadTopology() const 
{
    return (FD_ISSET(TRIP_ITAD_TOPOLOGY, &_aFlagSet));
}

bool 
TRIPUpdateMsg::hasAuthAttr() const 
{
    return (FD_ISSET(TRIP_AUTHENTICATION, &_aFlagSet));
}

Sptr<TRIPRouteAttr>
TRIPUpdateMsg::getAttribute(TRIPAttrType aType) const
{
    Sptr<TRIPRouteAttr> retAttr = 0;
    AttrSet::const_iterator itr;

    if(!FD_ISSET(aType, &_aFlagSet))
    {
        return retAttr;
    }
     
    for(itr = _attrSet.begin(); itr != _attrSet.end(); itr++)
    {
        const Sptr<TRIPRouteAttr> attr = (*itr);
        if(attr->getType() == aType)
        {
            return attr;
        }
    }
    return retAttr;
}

bool
TRIPUpdateMsg::validate(u_int16_t itad, u_int8_t &eCode, u_int8_t& subCode, string& msg) const
{
    eCode = TRIP_ERR_UPDATE; 
    ///Check for mandatory attributes
    try
    {
        checkSanity();
        //Check to see if both reachable and withdrwan route are link-state
        //or not.If withdrwn route is link-state the reachable route MUST
        //be link-state and vice-versa.
        if((_withdrawnRoute->isALinkStateAttr() ^ 
           _reachableRoute->isALinkStateAttr()) || 
           !checkAttrFlags(msg) )
        {
            cpLog(LOG_DEBUG, "Attr flags failed");
            subCode = U_ATTR_FLAG_ERR;
            return false;
        }
        if(!checkAttrLengths(msg))
        {
            cpLog(LOG_DEBUG, "Attr length check failed");
            subCode = U_ATTR_LENGTH_ERR;
            return false;
        }

        for(AttrSet::iterator itr = _attrSet.begin();
             itr != _attrSet.end(); itr++)
        {
            //If attribute does not get validated, throws a 
            //TRIPBadDataException with error code and subcode set in the
            //exception.
            (*itr)->validate();
        }

        //Check for itad loop in Advertisement path
        Sptr<TRIPRouteAttr> rAttr = getAttribute(TRIP_ADVERTISEMENT_PATH);
        Sptr<TRIPAdvertisePath> pAttr;
        pAttr.dynamicCast(rAttr);
        assert(pAttr != 0);
        if(pAttr->checkItadLoop(itad))
        {
            cpLog(LOG_DEBUG, "Attr check ItadLoop failed");
            subCode = U_INVALID_ATTR;
            return false;
        }
    }
    catch(TRIPBadDataException& e)
    {
        subCode = e.getErrSubcode();
        msg = e.getMsg();
        return false;
    }

    return true;
}

bool
TRIPUpdateMsg::checkAttrFlags(string& msg) const
{
    TRIPAttrFlag validReachWithdrawnFlg = 
          (TRIPAttrFlag)(TRIP_FLG_OPT | TRIP_FLG_TRANSITIVE | TRIP_FLG_DEPENDENT| TRIP_FLG_PARTIAL);
    if(_withdrawnRoute->getAttrFlags() & validReachWithdrawnFlg)
    {
        strstream cStr;
        cStr << *_withdrawnRoute << ends;
        msg  = cStr.str();
        cStr.freeze(false);
        return false;
    }
    if(_reachableRoute->getAttrFlags() & validReachWithdrawnFlg)
    {
        strstream cStr;
        cStr << *_reachableRoute << ends;
        msg  = cStr.str();
        cStr.freeze(false);
        return false;
    }
    for(AttrSet::iterator itr = _attrSet.begin();
             itr != _attrSet.end(); itr++)
    {
        bool failed = false;
        switch((*itr)->getType())
        {
            case TRIP_NEXT_HOP_SERVER:
            case TRIP_ATOMIC_AGGREGATE:
            case TRIP_LOCAL_PREF:
            case TRIP_MULTI_EXIT_DISC:
            {
                //Only well-known
                if((*itr)->getAttrFlags() & TRIP_FLG_ALL)
                {
                    failed = true;
                }
            }
            break;
            case TRIP_ADVERTISEMENT_PATH:
            case TRIP_ROUTED_PATH:
            case TRIP_AUTHENTICATION:
            {
                //Well-known or Partial
                if((*itr)->getAttrFlags() & (TRIP_FLG_ALL ^ TRIP_FLG_PARTIAL))
                {
                    failed = true;
                }
            }
            break;
            case TRIP_COMMUNITY:
            {
                //Optional , Idependent transitive
                u_int8_t aFlg = (*itr)->getAttrFlags(); 
                if((!(aFlg & (TRIP_FLG_OPT | TRIP_FLG_TRANSITIVE))) ||
                   (aFlg & TRIP_FLG_DEPENDENT) ||
                   (aFlg & (TRIP_FLG_PARTIAL | TRIP_FLG_LINK_STATE)))
                {
                    failed = true;
                }
            }
            break;
            case TRIP_ITAD_TOPOLOGY:
            {
                //Well-known and Link-State
                u_int8_t aFlg = (*itr)->getAttrFlags(); 
                if((aFlg &  (TRIP_FLG_ALL ^ TRIP_FLG_LINK_STATE)) ||
                   (!(aFlg & TRIP_FLG_LINK_STATE)))
                {
                    failed = true;
                }
            }
            break;
            default:
            break;
        }
        if(failed)
        {
            strstream cStr;
            cStr << *(*itr) << ends;
            msg  = cStr.str();
            cStr.freeze(false);
            return (false);
        }
    }
    return true;     
}

bool 
TRIPUpdateMsg::checkAttrLengths(string& msg) const
{
    if(_withdrawnRoute->getLength()  < 4) 
    {
        strstream cStr;
        cStr << *_withdrawnRoute << ends;
        msg  = cStr.str();
        cStr.freeze(false);
        return false;
    }
    if(_reachableRoute->getLength()  < 4) 
    {
        strstream cStr;
        cStr << *_reachableRoute << ends;
        msg  = cStr.str();
        cStr.freeze(false);
        return false;
    }
    for(AttrSet::const_iterator itr = _attrSet.begin();
             itr != _attrSet.end(); itr++)
    {
        bool failed = false;
        switch((*itr)->getType())
        {
            case TRIP_NEXT_HOP_SERVER:
            {
               if((*itr)->getLength() < 8) failed = true;
            }
            break;
            case TRIP_ADVERTISEMENT_PATH:
            case TRIP_ROUTED_PATH:
            case TRIP_COMMUNITY:
            {
               if(((*itr)->getLength() < 4) ||
                  ((*itr)->getLength() % 4)) failed = true;
            }
            break;
            case TRIP_ATOMIC_AGGREGATE:
            {
               if(((*itr)->getLength() != 4)) failed = true;
            }
            break;
            case TRIP_LOCAL_PREF:
            case TRIP_MULTI_EXIT_DISC:
            {
               if(((*itr)->getLength() != 8)) failed = true;
            }
            break;
            case TRIP_ITAD_TOPOLOGY:
            {
                if(((*itr)->getHeader().getHdrLength() < 12) ||
                  ((*itr)->getLength() % 4)) failed = true;
            }
            break;
            case TRIP_AUTHENTICATION:
            {
               if(((*itr)->getLength() < 16)) failed = true;
            }
            break;
            default:
            break;
        }
        if(failed)
        {
            strstream cStr;
            cStr << *(*itr) << ends;
            msg  = cStr.str();
            cStr.freeze(false);
            return (false);
        }
    }
    return true;
}

bool 
TRIPUpdateMsg::checkLinkStateAttributes(u_int8_t &eCode, u_int8_t& subCode, string& msg)
{
    eCode = TRIP_ERR_UPDATE;
    subCode = U_INVALID_ATTR;
    if(!(_withdrawnRoute->isALinkStateAttr() && 
           _reachableRoute->isALinkStateAttr()))
    {
        return false;
    }  

    for(AttrSet::iterator itr = _attrSet.begin();
             itr != _attrSet.end(); itr++)
    {
        if((*itr)->getType() == TRIP_ITAD_TOPOLOGY)
        {
            if(!(*itr)->isALinkStateAttr())
            {
                return false;
            }
        }
    }
    return true;
}

bool 
TRIPUpdateMsg::extCheckInvalidAttributes(u_int8_t &eCode, u_int8_t& subCode, string& msg)
{
    eCode = TRIP_ERR_UPDATE;
    subCode = U_INVALID_ATTR;
    if((_withdrawnRoute->isALinkStateAttr() || 
           _reachableRoute->isALinkStateAttr()))
    {
        msg += "Link-state attributes";
        return true;
    }  

    for(AttrSet::iterator itr = _attrSet.begin();
             itr != _attrSet.end(); itr++)
    {
        if(((*itr)->getType() == TRIP_ITAD_TOPOLOGY) ||
            ((*itr)->getType() == TRIP_LOCAL_PREF))
        {
            msg += "Topology or Local-pref attribute";
            return true;
        }
    }
    return false;
}
