#ifndef LOCKHELPER_HXX_
#define LOCKHELPER_HXX_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const LockHelper_hxx_version 
= "$Id: LockHelper.hxx,v 1.9 2001/01/23 22:40:55 bko Exp $";

#include <cassert>
#include "VMutex.h"

/** LockHelper uses constructor to lock a mutex and destructor to unlock a mutex
 *  Usage: 
 * <PRE>
 *      class myClass {
 *         public:
 *             void     doCriticalWork( void );
 *         private:
 *             VMutex   aMutex_;
 *             xxx      criticalData_;
 *      }
 *
 *      void myClass::doCriticalWork()
 *      {
 *           LockHelper lockAid( this->aMutex_ );   // lock the mutex
 *           //  critical code section
 *      }    // on exit of this method, ~lockAid() is invoked and
 *           // unlock aMutex_
 * </PRE>
 */
class LockHelper
{
    public:
        /// constructor.  takes a mutex to lock
        LockHelper( VMutex& mutex ) : itsLock(mutex), locked_(false)
        {
            itsLock.lock();
            locked_ = true;
        }
        /// destructor
        virtual ~LockHelper( )
        {
            if(locked_)
            {
                itsLock.unlock();
            }
        }

        /// lock the mutex, if needed
        void lock()  
        { 
            if(!locked_)
            {
                itsLock.lock(); 
                locked_ = true; 
            }
        };

        /// unlock the mutex.
        void unlock()  { itsLock.unlock(); locked_ = false; };

    private:
	/// suppress default constructor
	LockHelper();
	/// suppress copy constructor
	LockHelper(const LockHelper&);
	/// suppress assignment
	LockHelper& operator=(const LockHelper&);

        VMutex& itsLock;
        bool   locked_;
};

// LOCKHELPER_HXX_
#endif
